"""ThreatConnect Threat Intelligence Module"""
# third-party
import inflect

from .mappings.filters import Filters
from .mappings.group.group import Group
from .mappings.group.group_types.adversary import Adversary
from .mappings.group.group_types.campaign import Campaign
from .mappings.group.group_types.document import Document
from .mappings.group.group_types.email import Email
from .mappings.group.group_types.event import Event
from .mappings.group.group_types.incident import Incident
from .mappings.group.group_types.intrusion_set import IntrusionSet
from .mappings.group.group_types.report import Report
from .mappings.group.group_types.signature import Signature
from .mappings.group.group_types.threat import Threat
from .mappings.indicator.indicator import Indicator, custom_indicator_class_factory
from .mappings.indicator.indicator_types.address import Address
from .mappings.indicator.indicator_types.email_address import EmailAddress
from .mappings.indicator.indicator_types.file import File
from .mappings.indicator.indicator_types.host import Host
from .mappings.indicator.indicator_types.url import URL
from .mappings.owner import Owner
from .mappings.tag import Tag
from .mappings.task import Task
from .mappings.victim import Victim

p = inflect.engine()

# import local modules for dynamic reference
module = __import__(__name__)


class ThreatIntelligence:
    """ThreatConnect Threat Intelligence Module"""

    def __init__(self, tcex):
        """Initialize Class properties."""
        self.tcex = tcex
        self._custom_indicator_classes = {}
        self._gen_indicator_class()

    def address(self, **kwargs):
        """Return an Address TI object.

        Args:
            ip (str, kwargs): [Required for Create] The IP value for this Indicator.
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            last_modified (str, kwargs): [Read-Only] The date timestamp the Indicator was last
                modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.

        Returns:
            obj: An instance of Address.
        """
        return Address(self.tcex, **kwargs)

    def url(self, **kwargs):
        """Create the URL TI object.

        Args:
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            text (str, kwargs): [Required for Create] The URL value for this Indicator.

        Return:
            obj: An instance of URL.
        """
        return URL(self.tcex, **kwargs)

    def email_address(self, **kwargs):
        """Create the Email Address TI object.

        Args:
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            address (str, kwargs): [Required for Create] The Email Address value for this Indicator.

        Return:
            obj: An instance of EmailAddress.
        """
        return EmailAddress(self.tcex, **kwargs)

    def file(self, **kwargs):
        """Create the File TI object.

        Args:
            owner (str, kwargs): The name for this Group. Default to default Org when not provided

        Return:
            obj: An instance of File.
        """
        return File(self.tcex, **kwargs)

    def host(self, **kwargs):
        """Create the Host TI object.

        Args:
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            hostname (str, kwargs): [Required for Create] The Host value for this Indicator.

        Return:
            obj: An instance of Host.
        """
        return Host(self.tcex, **kwargs)

    def filters(self):
        """Create a Filters TI object"""
        return Filters(self.tcex)

    def indicator(self, indicator_type=None, owner=None, **kwargs):
        """Return an TI object.

        Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            indicator_type (str): The indicator type.
            ip (str, kwargs): [address] The value for this Indicator.
            last_modified (str, kwargs): [Read-Only] The date timestamp the Indicator was last
                modified.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.

        Returns:
            obj: An instance of Indicator or specific indicator type.
        """
        if not indicator_type:
            return Indicator(self.tcex, owner=owner, **kwargs)

        indicator_type_map = {
            'address': Address,
            'email address': EmailAddress,
            'emailaddress': EmailAddress,
            'file': File,
            'host': Host,
            'url': URL,
        }

        indicator_type = indicator_type.lower()
        for custom_type in self._custom_indicator_classes:
            for a in dir(module):
                if a.lower() == custom_type.replace(' ', ''):
                    indicator_type_map[custom_type.lower()] = getattr(module, a)

        if indicator_type not in indicator_type_map:
            raise RuntimeError(f'Invalid indicator type "{indicator_type}" provided.')

        # update kwargs
        kwargs['owner'] = owner
        kwargs['tcex'] = self.tcex

        # return correct indicator object
        indicator_object = indicator_type_map.get(indicator_type)
        return indicator_object(**kwargs)

    def group(self, group_type=None, owner=None, **kwargs):
        """Create the Group TI object.

        Args:
            owner (str): The ThreatConnect owner name.
            group_type: The type of group object.
        """
        if not group_type:
            return Group(self.tcex, owner=owner, **kwargs)

        group_type_map = {
            'adversary': Adversary,
            'campaign': Campaign,
            'document': Document,
            'event': Event,
            'email': Email,
            'incident': Incident,
            'intrusion set': IntrusionSet,
            'report': Report,
            'signature': Signature,
            'threat': Threat,
        }

        # if "name" is not in kwargs
        if kwargs.get('name') is None:
            kwargs['name'] = None

        group_type = group_type.lower()
        if group_type not in group_type_map:
            raise RuntimeError(f'Invalid group type "{group_type}" provided.')

        # update kwargs
        kwargs['owner'] = owner
        kwargs['tcex'] = self.tcex

        # return correct group object
        group_object = group_type_map.get(group_type)
        return group_object(**kwargs)

    def adversary(self, **kwargs):
        """Create the Adversary TI object.

        Args:
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided

        Return:
            ti.Adversary: An instance of Adversary.
        """
        return Adversary(self.tcex, **kwargs)

    def campaign(self, **kwargs):
        """Create the Campaign TI object.

        Args:
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            first_seen (str, kwargs): The first seen datetime expression for this Group.

        Return:
            ti.Campaign: An instance of Campaign.
        """
        return Campaign(self.tcex, **kwargs)

    def document(self, **kwargs):
        """Create the Document TI object.

        Args:
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            file_name (str, kwargs): The name for the attached file for this Group.
            malware (bool, kwargs): If true the file is considered malware.
            password (bool, kwargs): If malware is true a password for the zip archive is required.

        Return:
            ti.Document: An instance of Document.
        """
        return Document(self.tcex, **kwargs)

    def email(self, **kwargs):
        """Create the Email TI object.

        Args:
            body (str): The body for this Email.
            from_addr (str, kwargs): The **from** address for this Email.
            header (str): The header for this Email.
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            subject (str): The subject for this Email.
            to (str, kwargs): The **to** address for this Email.

        Return:
            ti.Email: An instance of Email.
        """
        return Email(self.tcex, **kwargs)

    def event(self, **kwargs):
        """Create the Event TI object.

        Args:
            event_date (str, kwargs): The event "event date" datetime expression for this Group.
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            status (str, kwargs): The status for this Group.

        Return:
            ti.Event: An instance of Event.
        """
        return Event(self.tcex, **kwargs)

    def incident(self, **kwargs):
        """Create the Incident TI object.

        Args:
            event_date (str, kwargs): The incident event date expression for this Group.
            name (str, kwargs): [Required for Create] The name for this Group.
            status (str, kwargs): The status for this Group.

        Return:
            ti.Incident: An instance of Incident.
        """
        return Incident(self.tcex, **kwargs)

    def intrusion_set(self, **kwargs):
        """Create the Intrustion Set TI object.

        Args:
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided

        Return:
            ti.IntrusionSet: An instance of IntrusionSet.
        """
        return IntrusionSet(self.tcex, **kwargs)

    def report(self, **kwargs):
        """Create the Report TI object.

        Args:
            owner (str): The ThreatConnect owner name.
            name:
            **kwargs:
        """
        return Report(self.tcex, **kwargs)

    def signature(self, **kwargs):
        """Create the Signature TI object.

        Args:
            owner (str): The ThreatConnect owner name.
            name (str): The name for this Group.
            file_name (str): The name for the attached signature for this Group.
            file_type (str): The signature type for this Group.
            file_text (str): The signature content for this Group.
            **kwargs:

        Return:
            obj: An instance of Signature.
        """
        return Signature(self.tcex, **kwargs)

    def task(self, **kwargs):
        """Create the Task TI object.

        Args:
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            status (str, kwargs): Not started, In Progress, Completed, Waiting on Someone, Deferred
            due_date (str, kwargs): Converted to %Y-%m-%dT%H:%M:%SZ date format
            reminder_date (str, kwargs): Converted to %Y-%m-%dT%H:%M:%SZ date format
            escalation_date (str, kwargs): Converted to %Y-%m-%dT%H:%M:%SZ date format

        Return:
            obj: An instance of Task
        """
        return Task(self.tcex, **kwargs)

    def threat(self, **kwargs):
        """Create the Threat TI object.

        Args:
            owner (str): The ThreatConnect owner name.
            name:
            **kwargs:
        """
        return Threat(self.tcex, **kwargs)

    def victim(self, **kwargs):
        """Create the Victim TI object.

        Args:
            owner (str): The ThreatConnect owner name.
            name:
            **kwargs:

        """
        return Victim(self.tcex, **kwargs)

    def tag(self, name):
        """Create the Tag TI object."""
        return Tag(self.tcex, name)

    def owner(self):
        """Create the Owner object."""
        return Owner(self.tcex)

    def create_entity(self, entity, owner):
        """Given a Entity and a Owner, creates a indicator/group in ThreatConnect"""

        attributes = entity.pop('attribute', [])
        associations = entity.pop('associations', [])
        security_labels = entity.pop('securityLabel', [])
        tags = entity.pop('tag', [])
        entity_type = entity.pop('type', '').lower()
        file_content = None
        if entity_type in ['document', 'report']:
            file_content = entity.pop('file_content', None) or entity.pop('fileContent', None)
        try:
            ti = self.indicator(entity_type, owner, **entity)
            if entity.get('falsePositive'):
                ti.add_false_positive()
        except Exception:
            if entity_type in ['victim']:
                ti = self.victim(owner=owner, **entity)
            else:
                entity['name'] = entity.pop('summary', None)
                if entity_type in ['task']:
                    ti = self.task(owner=owner, **entity)
                else:
                    ti = self.group(entity_type, owner, **entity)
        r = ti.create()
        if entity_type in ['document', 'report']:
            ti.file_content(file_content)

        data = {'status_code': r.status_code}
        if r.ok:
            data.update(r.json().get('data', {}))
            data['main_type'] = ti.type
            data['sub_type'] = ti.api_sub_type
            data['api_type'] = ti.api_sub_type
            data['api_entity']: ti.api_entity
            data['api_branch']: ti.api_branch
            data['owner'] = owner
            data['attributes'] = []
            data['tags'] = []
            data['security_labels'] = []
            data['associations'] = []

        for attribute in attributes:
            r = ti.add_attribute(attribute.get('type'), attribute.get('value'))
            attribute_data = {'status_code': r.status_code}
            if r.ok:
                attribute_data.update(r.json().get('attribute', {}))
            data['attributes'].append(attribute_data)
        for tag in tags:
            r = ti.add_tag(tag)
            tag_response = {'status_code': r.status_code}
            data['tags'].append(tag_response)
        for label in security_labels:
            r = ti.add_label(label)
            label_response = {'status_code': r.status_code}
            data['security_labels'].append(label_response)
        for association in associations:
            association_target = self.indicator(
                association.pop('type', None), association.pop('owner', None), **association
            )
            if not association_target:
                association_target = self.group(
                    association.pop('type', None), association.pop('owner', None), **association
                )
            r = ti.add_association(association_target)
            association_response = {'status_code': r.status_code}
            if r.ok:
                association_response.update(r.json().get('association', {}))
            data['associations'].append(association_response)

        return data

    def create_entities(self, entities, owner):
        """Create a indicator/group in TC based on the given entity's

        Args:
            entities: The entity to create.
            owner: The owner of the entity (
        """
        responses = []
        for entity in entities:
            responses.append(self.create_entity(entity, owner))
        return responses

    def entities(self, tc_data, resource_type):
        """Yield an entity.

        Takes both a list of indicators/groups or a individual indicator/group response.

        example formats

        .. code-block:: javascript

            {
                "status": "Success",
                "data": {
                    "resultCount": 984240,
                    "address": [
                        {
                            "id": 4222035,
                            "ownerName": "System",
                            "dateAdded": "2019-03-28T10:32:05-04:00",
                            "lastModified": "2019-03-28T11:02:46-04:00",
                            "rating": 4,
                            "confidence": 90,
                            "threatAssessRating": 4,
                            "threatAssessConfidence": 90,
                            "webLink": "{host}/auth/indicators/details/address.xhtml?
                                        address=221.123.32.14",
                            "ip": "221.123.32.14"
                        },
                        {
                            "id": 4221517,
                            "ownerName": "System",
                            "dateAdded": "2018-11-05T14:24:54-05:00",
                            "lastModified": "2019-03-07T12:38:36-05:00",
                            "threatAssessRating": 0,
                            "threatAssessConfidence": 0,
                            "webLink": "{host}/auth/indicators/details/address.xhtml?
                                        address=221.123.32.12",
                            "ip": "221.123.32.12"
                        }
                    ]
                }
            }

        or:

        .. code-block:: javascript

            {
                "status": "Success",
                "data": {
                    "address": {
                        "id": 4222035,
                        "owner": {
                            "id": 1,
                            "name": "System",
                            "type": "Organization"
                        },
                        "dateAdded": "2019-03-28T10:32:05-04:00",
                        "lastModified": "2019-03-28T11:02:46-04:00",
                        "rating": 4,
                        "confidence": 90,
                        "threatAssessRating": 4,
                        "threatAssessConfidence": 90,
                        "webLink": "{host}/auth/indicators/details/address.xhtml?
                                    address=221.123.32.14",
                        "ip": "221.123.32.14"
                    }
                }
            }

        Args:
            tc_data: TC data to convert to a entity.
            resource_type: The type of TC data being provided.

        """
        if not isinstance(tc_data, list):
            tc_data = [tc_data]

        for d in tc_data:
            entity = {'id': d.get('id'), 'webLink': d.get('webLink')}
            values = []
            value = None
            keys = d.keys()
            if resource_type.lower() in map(str.lower, self.tcex.group_types):
                r = self.tcex.ti.group(group_type=resource_type, name=d.get('name'))
                value = d.get('name')
            elif resource_type.lower() in map(str.lower, self.tcex.indicator_types):
                r = self.tcex.ti.indicator(indicator_type=resource_type)
                r._set_unique_id(d)
                value = r.unique_id
            elif resource_type.lower() in ['victim']:
                r = self.tcex.ti.victim(d.get('name'))
                value = d.get('name')
            else:
                self.tcex.handle_error(925, ['type', 'entities', 'type', 'type', resource_type])

            if 'summary' in d:
                values.append(d.get('summary'))
            else:
                if resource_type.lower() in ['file']:
                    value = r.build_summary(d.get('md5'), d.get('sha1'), d.get('sha256'))
                values.append(r.fully_decode_uri(value))
            entity['value'] = ' : '.join(values)

            if r.is_group() or r.is_indicator():
                if 'owner' in d:
                    entity['ownerName'] = d['owner']['name']
                else:
                    entity['ownerName'] = d.get('ownerName')
                entity['dateAdded'] = d.get('dateAdded')

            if r.is_victim():
                entity['ownerName'] = d.get('org')

            if r.is_indicator():
                entity['confidence'] = d.get('confidence')
                entity['rating'] = d.get('rating')
                entity['threatAssessConfidence'] = d.get('threatAssessConfidence')
                entity['threatAssessRating'] = d.get('threatAssessRating')
                entity['dateLastModified'] = d.get('lastModified')
                if 'whoisActive' in keys:
                    entity['whoisActive'] = d.get('whoisActive')
                if 'dnsActive' in keys:
                    entity['dnsActive'] = d.get('dnsActive')

            if r.is_task():
                entity['status'] = d.get('status')
                entity['escalated'] = d.get('escalated')
                entity['reminded'] = d.get('reminded')
                entity['overdue'] = d.get('overdue')
                entity['dueDate'] = d.get('dueDate', None)
                entity['reminderDate'] = d.get('reminderDate', None)
                entity['escalationDate'] = d.get('escalationDate', None)
                if d.get('xid'):
                    entity['xid'] = d.get('xid')
            if r.is_group():
                if 'xid' in keys:
                    entity['xid'] = d.get('xid')
                if 'firstSeen' in keys:
                    entity['firstSeen'] = d.get('firstSeen')
                if 'fileName' in keys:
                    entity['fileName'] = d.get('fileName')
                if 'fileType' in keys:
                    entity['fileType'] = d.get('fileType')
                if 'fileSize' in keys:
                    entity['fileSize'] = d.get('fileSize')
                if 'eventDate' in keys:
                    entity['eventDate'] = d.get('eventDate')
                if 'status' in keys:
                    entity['status'] = d.get('status')
                if 'to' in keys:
                    entity['to'] = d.get('to')
                if 'from' in keys:
                    entity['from'] = d.get('from')
                if 'subject' in keys:
                    entity['subject'] = d.get('subject')
                if 'score' in keys:
                    entity['score'] = d.get('score')
                if 'header' in keys:
                    entity['header'] = d.get('header')
                if 'body' in keys:
                    entity['body'] = d.get('body')
                if 'publishDate' in keys:
                    entity['publishDate'] = d.get('publishDate')
                if r.api_sub_type.lower() in ['signature', 'document', 'report']:
                    r.unique_id = d.get('id')
                    content_response = r.download()
                    if content_response.ok:
                        entity['fileContent'] = content_response.text
            # get the entity type
            if d.get('type') is not None:
                entity['type'] = d.get('type')
            else:
                entity['type'] = resource_type

            yield entity

    def _gen_indicator_class(self):
        """Generate Custom Indicator Classes."""

        for entry in self.tcex.indicator_types_data.values():
            name = entry.get('name')
            class_name = name.replace(' ', '')
            # temp fix for API issue where boolean are returned as strings
            entry['custom'] = self.tcex.utils.to_bool(entry.get('custom'))

            if class_name in globals():
                # skip Indicator Type if a class already exists
                continue

            # Custom Indicator can have 3 values. Only add the value if it is set.
            value_fields = []
            if entry.get('value1Label'):
                value_fields.append(entry['value1Label'])
            if entry.get('value2Label'):
                value_fields.append(entry['value2Label'])
            if entry.get('value3Label'):
                value_fields.append(entry['value3Label'])

            if not value_fields:
                continue

            # Add Class for each Custom Indicator type to this module
            custom_class = custom_indicator_class_factory(
                entry.get('name'),
                entry.get('apiEntity'),
                entry.get('apiBranch'),
                Indicator,
                value_fields,
            )

            custom_indicator_data = {
                'branch': entry.get('apiBranch'),
                'entry': entry.get('apiEntry'),
                'value_fields': value_fields,
            }
            self._custom_indicator_classes[entry.get('name').lower()] = custom_indicator_data

            setattr(module, class_name, custom_class)

            # Add Custom Indicator Method
            self._gen_indicator_method(name, custom_class)

    def _gen_indicator_method(self, name, custom_class):
        """Dynamically generate custom Indicator methods.

        Args:
            name (str): The name of the method.
            custom_class (object): The class to add.
        """
        method_name = name.replace(' ', '_').lower()
        tcex = self.tcex

        # Add Method for each Custom Indicator class
        def method_1(**kwargs):  # pylint: disable=possibly-unused-variable
            """Add Custom Indicator data to Batch object"""
            return custom_class(tcex, **kwargs)

        method = locals()['method_1']
        setattr(self, method_name, method)



"""ThreatConnect TI Generic Mappings Object"""
# standard library
import json
from urllib.parse import unquote

# first-party
from tcex.utils import Utils

from ..tcex_ti_tc_request import TiTcRequest


class Mappings:
    """Common API calls for for Indicators/SecurityLabels/Groups and Victims"""

    def __init__(self, tcex, main_type, api_type, sub_type, api_entity, api_branch, owner):
        """Initialize Class Properties.

        Args:
            tcex:
            main_type:
            api_type:
            sub_type:
            api_entity:
        """
        self._tcex = tcex
        self._data = {}

        self._owner = owner
        self._type = main_type
        self._api_sub_type = sub_type
        self._api_type = api_type
        self._unique_id = None
        self._api_entity = api_entity
        self._api_branch = api_branch

        self._utils = Utils()
        self._tc_requests = TiTcRequest(self._tcex)

    @property
    def _excluded_properties(self):
        """
        Returns a list of properties to exclude when creating a dict of the children Classes.
        """
        return ['tcex', 'kwargs', 'api_endpoint']

    @property
    def as_entity(self):  # pragma: no cover
        """Return the object as an entity."""
        raise NotImplementedError('Child class must implement this method.')

    @property
    def as_dict(self):
        """
        Returns the dict representation of the case management object.
        """
        properties = vars(self)
        as_dict = {}
        for key, value in properties.items():
            key = key.lstrip('_')
            if key in self._excluded_properties:
                continue
            try:
                value = value.as_dict
            except AttributeError:
                pass
            if value is None:
                continue
            as_dict[key] = value
        if not as_dict:
            return None
        return as_dict

    @property
    def type(self):
        """Return main type."""
        return self._type

    @property
    def api_sub_type(self):
        """Return sub type."""
        return self._api_sub_type

    @property
    def unique_id(self):
        """Return unique id."""
        return self._unique_id

    @property
    def tc_requests(self):
        """Return tc request object."""
        return self._tc_requests

    @property
    def api_type(self):
        """Return api type."""
        return self._api_type

    @property
    def api_branch(self):
        """Return api entity."""
        return self._api_branch

    @property
    def api_entity(self):
        """Return api entity."""
        return self._api_entity

    @property
    def owner(self):
        """Return unique id."""
        return self._owner

    @owner.setter
    def owner(self, owner):
        self._owner = owner

    @api_branch.setter
    def api_branch(self, api_branch):
        """
        Sets the Api Entity
        Args:
            api_branch:

        Returns:

        """
        self._api_branch = api_branch

    @api_entity.setter
    def api_entity(self, api_entity):
        """
        Sets the Api Entity
        Args:
            api_entity:

        Returns:

        """
        self._api_entity = api_entity

    @api_type.setter
    def api_type(self, api_type):
        """
        Sets the Api Type
        Args:
            api_type:

        Returns:

        """
        self._api_type = api_type

    @tc_requests.setter
    def tc_requests(self, tc_requests):
        """
        Sets the Tc Request Object
        Args:
            tc_requests:

        Returns:

        """
        self._tc_requests = tc_requests

    @api_sub_type.setter
    def api_sub_type(self, sub_type):
        """
        Sets the Api Sub Type
        Args:
            sub_type:

        Returns:

        """
        self._api_sub_type = sub_type

    @unique_id.setter
    def unique_id(self, unique_id):
        """
        Sets the Unique Id
        Args:
            unique_id:

        Returns:

        """
        self._unique_id = unique_id

    @property
    def data(self):
        """Return data."""
        return self._data

    @data.setter
    def data(self, data):
        """
        Sets the data
        Args:
            data:

        Returns:

        """
        self._data = data
        if not self.unique_id:
            self._set_unique_id(data)

    def set(self, **kwargs):
        """
        A generic way to update the attributes of a TC data object.
        Args:
            **kwargs:

        Returns:

        """
        for arg, value in kwargs.items():
            if hasattr(self, 'add_key_value'):
                self.add_key_value(arg, value)  # pylint: disable=no-member
            else:
                self._data[arg] = value

    def create(self):
        """
        Creates the Indicator/Group/Victim or Security Label given Owner

        Args:
        """
        if not self.can_create():
            self._tcex.handle_error(920, [self.type])

        response = self.tc_requests.create(self.api_type, self.api_branch, self._data, self.owner)

        if self.tc_requests.success(response):
            self._set_unique_id(response.json().get('data').get(self.api_entity))

        return response

    def delete(self):
        """
        Deletes the Indicator/Group/Victim or Security Label
        """
        if not self.can_delete():
            self._tcex.handle_error(915, [self.type])

        return self.tc_requests.delete(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )

    def update(self):
        """
        Updates the Indicator/Group/Victim or Security Label
        """
        if not self.can_update():
            self._tcex.handle_error(905, [self.type])

        return self.tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, self._data, owner=self.owner
        )

    def single(self, filters=None, params=None):
        """
        Gets the Indicator/Group/Victim or Security Label
        Args:
            filters:
            params: parameters to pass in to get the object

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.single(
            self.api_type,
            self.api_branch,
            self.unique_id,
            filters=filters,
            owner=self.owner,
            params=params,
        )

    def many(self, filters=None, params=None):
        """
        Gets the Indicator/Group/Victim or Security Labels
        Args:
            filters:
            owner:
            params: parameters to pass in to get the objects

        Yields: A Indicator/Group/Victim json

        """
        yield from self.tc_requests.many(
            self.api_type,
            self.api_branch,
            self.api_entity,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    def request(self, result_limit, result_start, filters=None, params=None):
        """
        Gets the Indicator/Group/Victim or Security Labels
        Args:
            filters:
            owner:
            result_limit:
            result_start:
            params: parameters to pass in to get the objects

        Returns:

        """
        return self.tc_requests.request(
            self.api_type,
            self.api_branch,
            result_limit,
            result_start,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    def tags(self, filters=None, params=None):
        """
         Gets the tags from a Indicator/Group/Victim/Security Labels
         Args:
             filters:
             owner:
             params: parameters to pass in to get the objects

         Yields: A tag json

         """

        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        yield from self.tc_requests.tags(
            self.api_type,
            self.api_branch,
            self.unique_id,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    def tag(self, name, action='ADD', params=None):
        """
         Adds a tag to a Indicator/Group/Victim/Security Label
         Args:
             params:
             action:
             name: The name of the tag

         """
        if not name:
            self._tcex.handle_error(925, ['name', 'tag', 'name', 'name', name])

        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if action in ['GET', 'ADD', 'DELETE']:
            return self.tc_requests.tag(
                self.api_type,
                self.api_branch,
                self.unique_id,
                name,
                action=action,
                owner=self.owner,
                params=params,
            )
        self._tcex.handle_error(925, ['action', 'tag', 'action', 'action', action])
        return None

    def add_tag(self, name):
        """
         Adds a tag to a Indicator/Group/Victim/Security Label
         Args:
             name: The name of the tag

         """
        return self.tag(name, action='ADD')

    def get_tag(self, name, params=None):
        """
         Gets a tag from a Indicator/Group/Victim/Security Label
         Args:
             name: The name of the tag
             params:
         """
        return self.tag(name, action='GET', params=params)

    def delete_tag(self, name):
        """
         Deletes a tag from a Indicator/Group/Victim/Security Label
         Args:
             name: The name of the tag
         """

        return self.tag(name, action='DELETE')

    def labels(self, filters=None, params=None):
        """
         Gets the security labels from a Indicator/Group/Victim

         Yields: A Security label

         """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        yield from self.tc_requests.labels(
            self.api_type,
            self.api_branch,
            self.unique_id,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    def label(self, label, action='ADD', params=None):
        """
         Adds a Security Label to a Indicator/Group or Victim
         Args:
             params:
             label: The name of the Security Label
             action:

         """

        if params is None:
            params = {}

        if not label:
            self._tcex.handle_error(925, ['label', 'Security Label', 'label', 'label', label])

        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if action == 'GET':
            return self.tc_requests.get_label(
                self.api_type,
                self.api_branch,
                self.unique_id,
                label,
                owner=self.owner,
                params=params,
            )

        if action == 'ADD':
            return self.tc_requests.add_label(
                self.api_type, self.api_branch, self.unique_id, label, owner=self.owner
            )

        if action == 'DELETE':
            return self.tc_requests.delete_label(
                self.api_type, self.api_branch, self.unique_id, label, owner=self.owner
            )

        self._tcex.handle_error(925, ['action', 'label', 'action', 'action', action])
        return None

    def add_label(self, label):
        """
         Adds a label to a Indicator/Group/Victim
         Args:
             label: The name of the Security Label
         """
        return self.label(label, action='ADD')

    def get_label(self, label, params=None):
        """
         Gets a security label from a Indicator/Group/Victim
         Args:
             label: The name of the Security Label
             params:
         """
        return self.label(label, action='GET', params=params)

    def delete_label(self, label):
        """
         Deletes a security label from a Indicator/Group/Victim
         Args:
             label: The name of the Security Label
         """
        return self.label(label, action='DELETE')

    def indicator_associations(self, params=None):
        """
         Gets the indicator association from a Indicator/Group/Victim

         Yields: Indicator Association

         """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if params is None:
            params = {}

        yield from self.tc_requests.indicator_associations(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner, params=params
        )

    def group_associations(self, params=None):
        """
         Gets the group association from a Indicator/Group/Victim

         Yields: Group Association

         """
        if params is None:
            params = {}

        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        yield from self.tc_requests.group_associations(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner, params=params
        )

    def victim_asset_associations(self, params=None):
        """
         Gets the victim asset association from a Indicator/Group/Victim

         Yields: Victim Association json

         """
        if params is None:
            params = {}
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.victim_asset_associations(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner, params=params
        )

    def indicator_associations_types(
        self, indicator_type, api_entity=None, api_branch=None, params=None
    ):
        """
        Gets the indicator association from a Indicator/Group/Victim

        Args:
            indicator_type:
            api_entity:
            api_branch:
            params:

        Returns:

        """
        if params is None:
            params = {}
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        target = self._tcex.ti.indicator(indicator_type)
        yield from self.tc_requests.indicator_associations_types(
            self.api_type,
            self.api_branch,
            self.unique_id,
            target,
            api_entity=api_entity,
            api_branch=api_branch,
            owner=self.owner,
            params=params,
        )

    def group_associations_types(self, group_type, api_entity=None, api_branch=None, params=None):
        """
        Gets the group association from a Indicator/Group/Victim

        Args:
            group_type:
            api_entity:
            api_branch:
            params:

        Returns:

        """
        if params is None:
            params = {}
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        target = self._tcex.ti.group(group_type)

        yield from self.tc_requests.group_associations_types(
            self.api_type,
            self.api_branch,
            self.unique_id,
            target,
            api_entity=api_entity,
            api_branch=api_branch,
            owner=self.owner,
            params=params,
        )

    def victim_asset_associations_type(self, victim_asset_type, params=None):
        """
        Gets the victim association from a Indicator/Group/Victim

        Args:
            victim_asset_type:
            params:

        Returns:

        """
        if params is None:
            params = {}
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.victim_asset_associations(
            self.api_type,
            self.api_branch,
            self.unique_id,
            victim_asset_type,
            owner=self.owner,
            params=params,
        )

    def add_association(self, target, api_type=None, api_branch=None, unique_id=None):
        """
        Adds a association to a Indicator/Group/Victim

        Args:
            target:
            api_type:
            api_branch:
            unique_id:

        Returns:

        """
        api_type = api_type or target.api_type
        api_branch = api_branch or target.api_branch
        unique_id = unique_id or target.unique_id
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if not target.can_update():
            self._tcex.handle_error(910, [target.type])

        return self.tc_requests.add_association(
            self.api_type,
            self.api_branch,
            self.unique_id,
            api_type,
            api_branch,
            unique_id,
            owner=self.owner,
        )

    def delete_association(self, target, api_type=None, api_branch=None, unique_id=None):
        """
        Deletes a association from a Indicator/Group/Victim

        Args:
            target:
            api_type:
            api_branch:
            unique_id:

        Returns:

        """
        api_type = api_type or target.api_type
        api_branch = api_branch or target.api_branch
        unique_id = unique_id or target.unique_id
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if not target.can_update():
            self._tcex.handle_error(910, [target.type])

        return self.tc_requests.delete_association(
            self.api_type,
            self.api_branch,
            self.unique_id,
            api_type,
            api_branch,
            unique_id,
            owner=self.owner,
        )

    def attributes(self, params=None):
        """
        Gets the attributes from a Group/Indicator or Victim

        Yields: attribute json

        """
        if params is None:
            params = {}
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        yield from self.tc_requests.attributes(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner, params=params
        )

    def attribute(self, attribute_id, action='GET', params=None):
        """
        Gets the attribute from a Group/Indicator or Victim


        Args:
            action:
            params:
            attribute_id:

        Returns: attribute json

        """
        if params is None:
            params = {}
        action = action.upper()
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if action == 'GET':
            return self.tc_requests.get_attribute(
                self.api_type,
                self.api_branch,
                self.unique_id,
                attribute_id,
                owner=self.owner,
                params=params,
            )

        if action == 'DELETE':
            return self.tc_requests.delete_attribute(
                self.api_type, self.api_branch, self.unique_id, attribute_id, owner=self.owner
            )

        self._tcex.handle_error(925, ['action', 'attribute', 'action', 'action', action])
        return None

    def add_attribute(
        self, attribute_type, attribute_value, source=None, displayed=None, params=None
    ):
        """
        Adds a attribute to a Group/Indicator or Victim


        Args:
            source:
            displayed:
            params:
            attribute_type:
            attribute_value:

        Returns: attribute json

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if params is None:
            params = {}

        return self.tc_requests.add_attribute(
            self.api_type,
            self.api_branch,
            self.unique_id,
            attribute_type,
            attribute_value,
            source=source,
            displayed=displayed,
            owner=self.owner,
            params=params,
        )

    def update_attribute(
        self, attribute_value, attribute_id, source=None, displayed=None, params=None
    ):
        """
        Adds a attribute to a Group/Indicator or Victim


        Args:
            source:
            displayed:
            params:
            attribute_type:
            attribute_value:

        Returns: attribute json

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if params is None:
            params = {}

        return self.tc_requests.update_attribute(
            self.api_type,
            self.api_branch,
            self.unique_id,
            attribute_value,
            attribute_id,
            source=source,
            displayed=displayed,
            owner=self.owner,
            params=params,
        )

    def delete_attribute(self, attribute_id):
        """
        Deletes a attribute from a Group/Indicator or Victim


        Args:
            attribute_id:

        Returns: attribute json

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.delete_attribute(
            self.api_type, self.api_branch, self.unique_id, attribute_id, owner=self.owner
        )

    def attribute_labels(self, attribute_id, params=None):
        """
        Gets the security labels from a attribute

        Yields: Security label json

        """
        if params is None:
            params = {}
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        yield from self.tc_requests.attribute_labels(
            self.api_type,
            self.api_branch,
            self.unique_id,
            attribute_id,
            owner=self.owner,
            params=params,
        )

    def attribute_label(self, attribute_id, label, action='GET', params=None):
        """
        Gets a security labels from a attribute

        Args:
            attribute_id:
            label:
            action:
            params:

        Returns: Security label json
        """
        if params is None:
            params = {}
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if action == 'GET':
            return self.tc_requests.get_attribute_label(
                self.api_type,
                self.api_branch,
                self.unique_id,
                attribute_id,
                label,
                owner=self.owner,
                params=params,
            )
        if action == 'DELETE':
            return self.tc_requests.delete_attribute_label(
                self.api_type,
                self.api_branch,
                self.unique_id,
                attribute_id,
                label,
                owner=self.owner,
            )

        self._tcex.handle_error(925, ['action', 'attribute_label', 'action', 'action', action])
        return None

    def add_attribute_label(self, attribute_id, label):
        """
        Adds a security labels to a attribute

        Args:
            attribute_id:
            label:

        Returns: A response json
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.add_attribute_label(
            self.api_type, self.api_branch, self.unique_id, attribute_id, label, owner=self.owner
        )

    def can_create(self):  # pylint: disable=no-self-use
        """ Determines if the object can be created. """
        return True

    def can_delete(self):
        """ Determines if the object can be deleted. """
        return self.unique_id is not None

    def can_update(self):
        """ Determines if the object can be updated. """
        return self.unique_id is not None

    @staticmethod
    def is_indicator():
        """ Determines if the object is a Indicator. """
        return False

    @staticmethod
    def is_group():
        """ Determines if the object is a Group. """
        return False

    @staticmethod
    def is_victim():
        """ Determines if the object is a Victim. """
        return False

    @staticmethod
    def is_tag():
        """ Determines if the object is a Tag. """
        return False

    @staticmethod
    def is_security_label():
        """ Determines if the object is a Security Label. """
        return False

    @staticmethod
    def is_task():
        """ Determines if the object is a Task. """
        return False

    @staticmethod
    def is_encoded(uri):
        """Determines if a uri is currently encoded"""
        uri = uri or ''

        return uri != unquote(uri)

    def fully_decode_uri(self, uri):
        """Decodes a url till it is no longer encoded."""
        saftey_valve = 0

        while self.is_encoded(uri):
            uri = unquote(uri)
            saftey_valve += 1
            if saftey_valve > 10:
                break

        return uri

    def _set_unique_id(self, json_response):
        """ Sets the Unique Id given a json """
        self.unique_id = json_response.get('id')

    def __str__(self):
        """Return string representation of object."""
        return json.dumps(self.data, indent=4)



"""ThreatConnect Threat Intelligence Module"""
# standard library
import hashlib
from urllib.parse import quote

# import local modules for dynamic reference
module = __import__(__name__)


class TiTcRequest:
    """Common API calls to ThreatConnect

    Args:
        tcex ([type]): [description]
    """

    def __init__(self, tcex):
        """Initialize Class properties."""
        self.tcex = tcex
        self.result_limit = 10000

    def _delete(self, url, params=None):
        """Delete data from API."""
        params = params or {}
        params['createActivityLog'] = params.get('createActivityLog') or 'false'

        r = self.tcex.session.delete(url, params=params)
        self.tcex.log.debug(
            f'Method: ({r.request.method.upper()}), '
            f'Params: ({params}), '
            f'Status Code: {r.status_code}, '
            f'URL: ({r.url})'
        )
        if len(r.content) < 500:
            self.tcex.log.trace(f'response: {r.text}')
        if not r.ok:
            err = r.text or r.reason
            self.tcex.log.error(f'Error deleting data ({err}')
        return r

    def _get(self, url, params=None):
        """Delete data from API."""
        params = params or {}
        params['createActivityLog'] = params.get('createActivityLog') or 'false'

        r = self.tcex.session.get(url, params=params)

        self.tcex.log.debug(
            f'Method: ({r.request.method.upper()}), '
            f'Params: ({params}), '
            f'Status Code: {r.status_code}, '
            f'URL: ({r.url})'
        )
        if len(r.content) < 500:
            self.tcex.log.trace(f'response: {r.text}')
        if not r.ok:
            err = r.text or r.reason
            self.tcex.log.error(f'Error getting data ({err}')
        return r

    def _iterate(self, url, params, api_entity):
        """Iterate over API pagination."""
        params['resultLimit'] = self.result_limit

        should_iterate = True
        result_start = params.get('resultStart', 0)
        try:
            result_start = int(result_start)
        except Exception:
            result_start = 0
            self.tcex.log.error('Invalid ResultStart Param. Starting at 0')
        while should_iterate:
            params['resultStart'] = result_start
            r = self._get(url, params=params)
            if not self.success(r):
                err = r.text or r.reason
                self.tcex.handle_error(950, [r.status_code, err, r.url])
            data = r.json().get('data', {})
            if api_entity:
                data = data.get(api_entity, [])

            if len(data) < self.result_limit:
                should_iterate = False
            result_start += self.result_limit

            yield from data

    def _post(self, url, data, params=None):
        """Post data to API."""
        params = params or {}
        params['createActivityLog'] = params.get('createActivityLog') or 'false'

        r = self.tcex.session.post(url, data=data, params=params)
        self.tcex.log.debug(
            f'Method: ({r.request.method.upper()}), '
            f'Params: ({params}), '
            f'Status Code: {r.status_code}, '
            f'URL: ({r.url})'
        )
        if len(data) < 50 and not isinstance(data, bytes):
            self.tcex.log.trace(f'body: {data}')
        if len(r.content) < 500:
            self.tcex.log.trace(f'response: {r.text}')
        if not r.ok:
            err = r.text or r.reason
            self.tcex.log.error(f'Error posting data ({err}')
        return r

    def _post_json(self, url, json_data, params=None):
        """Post JSON data to API."""
        params = params or {}
        params['createActivityLog'] = params.get('createActivityLog') or 'false'

        r = self.tcex.session.post(url, json=json_data, params=params)
        self.tcex.log.debug(
            f'Method: ({r.request.method.upper()}), '
            f'Params: ({params}), '
            f'Status Code: {r.status_code}, '
            f'URL: ({r.url})'
        )
        self.tcex.log.trace(f'body: {json_data}')
        if len(r.content) < 500:
            self.tcex.log.trace(f'response: {r.text}')
        if not r.ok:
            err = r.text or r.reason
            self.tcex.log.error(f'Error posting data ({err}')
        return r

    def _put_json(self, url, json_data, params=None):
        """Post JSON data to API."""
        params = params or {}
        params['createActivityLog'] = params.get('createActivityLog') or 'false'

        r = self.tcex.session.put(url, json=json_data, params=params)
        self.tcex.log.debug(
            f'Method: ({r.request.method.upper()}), '
            f'Params: ({params}), '
            f'Status Code: {r.status_code}, '
            f'URL: ({r.url})'
        )
        if len(json_data) < 50 and not isinstance(json_data, bytes):
            self.tcex.log.trace(f'body: {json_data}')
        if len(r.content) < 500:
            self.tcex.log.trace(f'response: {r.text}')
        if not r.ok:
            err = r.text or r.reason
            self.tcex.log.error(f'Error updating data ({err}')
        return r

    def add_adversary_handle_asset(self, unique_id, value):
        """Add an asset to an Adversary.

        Args:
            unique_id (str): The unique ID of the Adversary
            value (str): The asset value

        Returns:
            requests.Response: A request Response object.
        """
        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/handles'
        asset_data = {'handle': value}
        return self._post_json(asset_url, asset_data)

    def add_adversary_phone_asset(self, unique_id, value):
        """Add an asset to an Adversary.

        Args:
            unique_id (str): The unique ID of the Adversary
            value (str): The asset value

        Returns:
            requests.Response: A request Response object.
        """
        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/phoneNumbers'
        asset_data = {'phoneNumber': value}
        return self._post_json(asset_url, asset_data)

    def add_adversary_url_asset(self, unique_id, value):
        """Add an asset to an Adversary.

        Args:
            unique_id (str): The unique ID of the Adversary
            value (str): The asset value

        Returns:
            requests.Response: A request Response object.
        """
        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/urls'
        asset_data = {'url': value}
        return self._post_json(asset_url, asset_data)

    def adversary_assets(self, unique_id, params=None):
        """Return all Adversary assets

        Args:
            unique_id (str): The unique ID of the Adversary
            params (dict, optional): The query params for the request. Defaults to None.

        Yields:
            requests.Response: A request Response object.
        """
        params = params or {}

        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets'
        yield from self._iterate(asset_url, params, 'bucketAsset')

    def adversary_handle_asset(self, unique_id, asset_id, action='GET', params=None):
        """Return Adversary handle asset by ID

        Args:
            unique_id (str): The unique ID of the Adversary
            asset_id: (str) The ID of the asset.
            action: (str): The HTTP method (e.g., DELETE or GET)
            params (dict, optional): The query params for the request. Defaults to None.

        Returns:
            requests.Response: A request Response object.
        """
        params = params or {}

        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/handles/{asset_id}'

        # handle DELETE action
        if action == 'DELETE':
            return self._delete(asset_url, params=params)

        return self._get(asset_url, params=params)

    def adversary_handle_assets(self, unique_id, params=None):
        """Return all Adversary handle assets

        Args:
            unique_id (str): The unique ID of the Adversary
            params (dict, optional): The query params for the request. Defaults to None.

        Yields:
            requests.Response: A request Response object.
        """
        params = params or {}

        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/handles'
        yield from self._iterate(asset_url, params, 'adversaryHandle')

    def adversary_phone_asset(self, unique_id, asset_id, action='GET', params=None):
        """Return Adversary phone number asset by ID

        Args:
            unique_id (str): The unique ID of the Adversary
            asset_id: (str) The ID of the asset.
            action: (str): The HTTP method (e.g., DELETE or GET)
            params (dict, optional): The query params for the request. Defaults to None.

        Returns:
            requests.Response: A request Response object.
        """
        params = params or {}

        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/phoneNumbers/{asset_id}'

        # handle DELETE action
        if action == 'DELETE':
            return self._delete(asset_url, params=params)

        return self._get(asset_url, params=params)

    def adversary_phone_assets(self, unique_id, params=None):
        """Return all Adversary phone assets

        Args:
            unique_id (str): The unique ID of the Adversary
            params (dict, optional): The query params for the request. Defaults to None.

        Yields:
            requests.Response: A request Response object.
        """
        params = params or {}

        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/phoneNumbers'
        yield from self._iterate(asset_url, params, 'adversaryPhoneNumber')

    def adversary_url_asset(self, unique_id, asset_id, action='GET', params=None):
        """Return Adversary url asset by ID

        Args:
            unique_id (str): The unique ID of the Adversary
            asset_id: (str) The ID of the asset.
            action: (str): The HTTP method (e.g., DELETE or GET)
            params (dict, optional): The query params for the request. Defaults to None.

        Returns:
            requests.Response: A request Response object.
        """
        params = params or {}

        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/urls/{asset_id}'

        # handle DELETE action
        if action == 'DELETE':
            return self._delete(asset_url, params=params)

        return self._get(asset_url, params=params)

    def adversary_url_assets(self, unique_id, params=None):
        """Return all Adversary url assets

        Args:
            unique_id (str): The unique ID of the Adversary
            params (dict, optional): The query params for the request. Defaults to None.

        Yields:
            requests.Response: A request Response object.
        """
        params = params or {}

        asset_url = f'/v2/groups/adversaries/{unique_id}/adversaryAssets/urls'
        yield from self._iterate(asset_url, params, 'adversaryUrl')

    def create(self, main_type, sub_type, data, owner):
        """Create a TI object in the API.

        Args:
            main_type (str): The TI type (e.g., groups or indicators).
            sub_type (str): The TI sub type (e.g., adversaries or addresses).
            data (dict): The body for the POST.
            owner (str): The name of the TC owner.

        Returns:
            request.Response: The response from the API call.
        """
        url = f'/v2/{main_type}'
        if sub_type:
            url = f'/v2/{main_type}/{sub_type}'

        params = {}
        if owner:
            params['owner'] = owner
        return self._post_json(url, data, params)

    def delete(self, main_type, sub_type, unique_id, owner=None):
        """Delete a TI object in the API.

        Args:
            main_type (str): The TI type (e.g., groups or indicators).
            sub_type (str): The TI sub type (e.g., adversaries or addresses).
            data (dict): The body for the POST.
            owner (str): The name of the TC owner.

        Returns:
            request.Response: The response from the API call.
        """
        url = f'/v2/{main_type}/{unique_id}'
        if sub_type:
            if unique_id:
                url = f'/v2/{main_type}/{sub_type}/{unique_id}'

        params = {}
        if owner:
            params['owner'] = owner
        return self._delete(url, params)

    def delete_adversary_handle_asset(self, unique_id, asset_id):
        """Delete Adversary handle assest

        Args:
            unique_id (str): The unique ID of the Adversary.
            asset_id: (str) The ID of the asset.

        Returns:
            requests.Response: A request Response object.
        """
        return self.adversary_handle_asset(unique_id, asset_id, action='DELETE')

    def delete_adversary_phone_asset(self, unique_id, asset_id):
        """Delete Adversary phone assest

        Args:
            unique_id (str): The unique ID of the Adversary.
            asset_id: (str) The ID of the asset.

        Returns:
            requests.Response: A request Response object.
        """
        return self.adversary_phone_asset(unique_id, asset_id, action='DELETE')

    def delete_adversary_url_asset(self, unique_id, asset_id):
        """Delete Adversary URL assest

        Args:
            unique_id (str): The unique ID of the Adversary.
            asset_id: (str) The ID of the asset.

        Returns:
            requests.Response: A request Response object.
        """
        return self.adversary_url_asset(unique_id, asset_id, action='DELETE')

    def download(self, main_type, sub_type, unique_id):
        """Download a Document or Report

        Args:
            main_type (str): The TI type (e.g., groups or indicators).
            sub_type (str): The TI sub type (e.g., adversaries or addresses).
            unique_id (str): The unique ID of the Resource.

        Returns:
            requests.Response: A request Response object.
        """
        url = f'/v2/{main_type}/{unique_id}/download'
        if sub_type:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/download'
        return self._get(url)

    def get_adversary_handle_asset(self, unique_id, asset_id, params=None):
        """Get Adversary handle assest

        Args:
            unique_id (str): The unique ID of the Adversary.
            asset_id: (str) The ID of the asset.

        Returns:
            requests.Response: A request Response object.
        """
        return self.adversary_handle_asset(unique_id, asset_id, params=params)

    def get_adversary_phone_asset(self, unique_id, asset_id, params=None):
        """Get Adversary phone assest

        Args:
            unique_id (str): The unique ID of the Adversary.
            asset_id: (str) The ID of the asset.

        Returns:
            requests.Response: A request Response object.
        """
        return self.adversary_phone_asset(unique_id, asset_id, params=params)

    def get_adversary_url_asset(self, unique_id, asset_id, params=None):
        """Get Adversary URL assest

        Args:
            unique_id (str): The unique ID of the Adversary.
            asset_id: (str) The ID of the asset.

        Returns:
            requests.Response: A request Response object.
        """
        params = params or {}

        return self.adversary_url_asset(unique_id, asset_id, params=params)

    def many(self, main_type, sub_type, api_entity, owner=None, filters=None, params=None):
        """Update a TI object in the API.

        Args:
            main_type (str): The TI type (e.g., groups or indicators).
            sub_type (str): The TI sub type (e.g., adversaries or addresses).
            api_entity (str): The API entity value (e.g., address, file, etc).
            owner (str): The name of the TC owner.
            filters (Filter, optional): A filter object.
            params (dict, optional): Optional dict of query params.

        Yields:
            request.Response: The response from the API call.
        """
        params = params or {}

        # add owner
        if owner:
            params['owner'] = owner

        # add filters
        if filters and filters.filters:
            params['filters'] = filters.filters_string

        # build url
        url = f'/v2/{main_type}'
        if sub_type:
            url = f'/v2/{main_type}/{sub_type}'

        yield from self._iterate(url, params, api_entity)

    def mine(self):
        """Get owner mine data."""
        return self._get('/v2/owners/mine')

    def request(
        self, main_type, sub_type, result_limit, result_start, owner=None, filters=None, params=None
    ):
        """[summary]

        Args:
            main_type ([type]): [description]
            sub_type ([type]): [description]
            result_limit ([type]): [description]
            result_start ([type]): [description]
            owner ([type], optional): [description]. Defaults to None.
            filters ([type], optional): [description]. Defaults to None.
            params ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """
        params = params or {}

        # add owner
        if owner:
            params['owner'] = owner

        # add filters
        if filters and filters.filters:
            params['filters'] = filters.filters_string

        # add result limits and result start
        params['resultLimit'] = result_limit or params.get('result_limit', self.result_limit)
        params['resultStart'] = result_start or params.get('result_start', 0)

        url = f'/v2/{main_type}'
        if sub_type:
            url = f'/v2/{main_type}/{sub_type}'

        return self._get(url, params)

    def single(self, main_type, sub_type, unique_id, owner=None, filters=None, params=None):
        """Update a TI object in the API.

        Args:
            main_type (str): The TI type (e.g., groups or indicators).
            sub_type (str): The TI sub type (e.g., adversaries or addresses).
            unique_id (str): The unique ID of the TI object.
            owner (str): The name of the TC owner.
            filters (Filter, optional): A filter object.
            params (dict, optional): Optional dict of query params.

        Returns:
            request.Response: The response from the API call.
        """
        params = params or {}

        # add owner
        if owner:
            params['owner'] = owner

        # add filters
        if filters and filters.filters:
            params['filters'] = filters.filters_string

        # build url
        url = f'/v2/{main_type}/{unique_id}'
        if sub_type:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}'

        return self._get(url, params)

    def update(self, main_type, sub_type, unique_id, data, owner=None):
        """Update a TI object in the API.

        Args:
            main_type (str): The TI type (e.g., groups or indicators).
            sub_type (str): The TI sub type (e.g., adversaries or addresses).
            unique_id (str): The unique ID of the TI object.
            data (dict): The body for the POST.
            owner (str): The name of the TC owner.

        Returns:
            request.Response: The response from the API call.
        """
        url = f'/v2/{main_type}/{unique_id}'
        if sub_type:
            if unique_id:
                url = f'/v2/{main_type}/{sub_type}/{unique_id}'

        params = {}
        if owner:
            params['owner'] = owner
        return self._put_json(url, data, params)

    def upload(self, main_type, sub_type, unique_id, data, update_if_exists=True):
        """Upload a file to API.

        Args:
            main_type ([type]): [description]
            sub_type ([type]): [description]
            unique_id ([type]): [description]
            data ([type]): [description]
            update_if_exists (bool, optional): [description]. Defaults to True.

        Returns:
            [type]: [description]
        """
        if not isinstance(data, bytes):
            data = bytes(data, 'utf-8')

        params = {}
        if update_if_exists:
            params['updateIfExists'] = 'true'

        url = f'/v2/{main_type}/{sub_type}/{unique_id}/upload'
        return self._post(url, data=data, params=params)

    def victim_add_asset(self, unique_id, asset_type, body):
        """Add a Asset to a Victim.

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            unique_id (str): The Victims ID.
            asset_type: (str) The type of asset to be retrieved.
            body (dict): The body of the Asset.

        Returns:
            request.Response: The response from the API call.
        """
        asset_type = self.victim_asset_type_mapping.get(asset_type.lower(), asset_type)
        asset_url = f'/v2/victims/{unique_id}/victimAssets/{asset_type}'
        return self._post_json(asset_url, body)

    def victim_assets(self, main_type, sub_type, unique_id, asset_type=None, params=None):
        """Retrieve the Assets of a Victim.

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            main_type (str): The main type of the TI object.
            sub_type (str): The sub type of the TI object.
            unique_id (str): The unique id of the TI object.
            asset_type: (str) The type of asset to be retrieved. Defaults to all of them.
            params (dict, optional): Optional dict of query params.

        Yields:
            Json: The asset being retrieved.
        """
        params = params or {}

        # This is needed because of tasks
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/victimAssets'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/victimAssets'
        entity_type = 'victimAsset'
        asset_type = self.victim_asset_type_mapping.get(asset_type)
        if asset_type:
            url += f'/{asset_type}'
        return self._iterate(url, params, entity_type)

    def victim_get_asset(self, unique_id, asset_type, asset_id, params=None):
        """Retrieve a Asset from a Victim.

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            unique_id (str): The Victims ID.
            asset_type: (str) The type of asset to be retrieved.
            asset_id (str): The Assets ID .
            params (dict, optional): Optional dict of query params.

        Returns:
            request.Response: The response from the API call.
        """
        params = params or {}
        asset_type = self.victim_asset_type_mapping.get(asset_type.lower(), asset_type)
        asset_url = f'/v2/victims/{unique_id}/victimAssets/{asset_type}/{asset_id}'
        return self._get(asset_url, params=params)

    @property
    def victim_asset_type_mapping(self):
        """Maps the asset type to the api endpoints needed."""
        return {
            'email': 'emailAddresses',
            'network': 'networkAccounts',
            'phone': 'phoneNumbers',
            'social': 'socialNetworks',
            'web': 'webSites',
        }

    def victim_delete_asset(self, unique_id, asset_type, asset_id):
        """Delete a Asset from a Victim.

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            unique_id (str): The Victims ID.
            asset_type: (str) The type of asset to be retrieved.
            asset_id (str): The Assets ID .

        Returns:
            request.Response: The response from the API call.
        """
        asset_type = self.victim_asset_type_mapping.get(asset_type.lower(), asset_type)
        asset_url = f'/v2/victims/{unique_id}/victimAssets/{asset_type}/{asset_id}'
        return self._delete(asset_url)

    def victim_update_asset(self, unique_id, asset_type, asset_id, body):
        """Update a Asset of a Victim.

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            unique_id (str): The Victims ID.
            asset_type: (str) The type of asset to be retrieved.
            asset_id (str): The Assets ID .
            body (dict): The body of the Asset.

        Returns:
            request.Response: The response from the API call.
        """
        asset_type = self.victim_asset_type_mapping.get(asset_type.lower(), asset_type)
        asset_url = f'/v2/victims/{unique_id}/victimAssets/{asset_type}/{asset_id}'
        return self._put_json(asset_url, body)

    #
    # -- needs updates ---
    #

    def add_false_positive(self, main_type, sub_type, unique_id, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:

        Return:

        """
        params = {'owner': owner} if owner else {}

        url = f'/v2/{main_type}/{sub_type}/{unique_id}/falsePositive'

        r = self._post(url, {}, params)
        self.tcex.log.debug(f'status code: {r.status_code}')
        self.tcex.log.trace(f'url: {r.request.url}')
        return r

    def owners(self, main_type, sub_type, unique_id, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:

        Return:

        """
        params = {'owner': owner} if owner else {}
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/owners'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/owners'

        r = self._get(url, params=params)
        self.tcex.log.debug(f'status code: {r.status_code}')
        self.tcex.log.trace(f'url: {r.request.url}')
        return r

    def add_observations(self, main_type, sub_type, unique_id, data, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            data:

        Return:

        """

        params = {'owner': owner} if owner else {}

        url = f'/v2/{main_type}/{sub_type}/{unique_id}/observations'
        r = self._post_json(url, data, params)
        self.tcex.log.debug(f'status code: {r.status_code}')
        self.tcex.log.trace(f'url: {r.request.url}')
        return r

    def observation_count(self, main_type, sub_type, unique_id, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:

        Return:

        """
        params = {'owner': owner} if owner else {}

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/observationCount'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/observationCount'

        r = self._get(url, params)
        self.tcex.log.debug(f'status code: {r.status_code}')
        self.tcex.log.trace(f'url: {r.request.url}')
        return r

    def observations(self, main_type, sub_type, unique_id, owner=None, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            owner:
            params:

        Return:

        """
        params = params or {}

        if owner:
            params['owner'] = owner

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/observations'
        else:
            url = f'/v2/{type}/{sub_type}/{unique_id}/observations'

        r = self._get(url, params)
        self.tcex.log.debug(f'status code: {r.status_code}')
        self.tcex.log.trace(f'url: {r.request.url}')
        return r

    def dns_resolution(self, main_type, sub_type, unique_id, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:

        Return:

        """
        params = {'owner': owner} if owner else {}

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/dnsResolution'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/dnsResolution'

        r = self._get(url, params)
        self.tcex.log.debug(f'status code: {r.status_code}')
        self.tcex.log.trace(f'url: {r.request.url}')
        return r

    def set_dns_resolution(self, main_type, sub_type, unique_id, value, owner=None):
        """

         Args:
             value:
             owner:
             main_type:
             sub_type:
             unique_id:

         Return:

         """
        params = {'owner': owner} if owner else {}

        data = {}
        if self.is_true(value) or self.is_false(value):
            data['dnsActive'] = self.is_true(value)
        else:
            self.tcex.handle_error(925, ['option', 'dns value', 'value', value])

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}'

        r = self._put_json(url, data, params)
        self.tcex.log.debug(f'status code: {r.status_code}')
        self.tcex.log.trace(f'url: {r.request.url}')
        return r

    def set_whois(self, main_type, sub_type, unique_id, value, owner=None):
        """

          Args:
              value:
              owner:
              main_type:
              sub_type:
              unique_id:

          Return:

          """
        params = {'owner': owner} if owner else {}

        data = {}
        if self.is_true(value) or self.is_false(value):
            data['whoisActive'] = self.is_true(value)
        else:
            self.tcex.handle_error(925, ['option', 'whois value', 'value', value])

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}'

        r = self._put_json(url, data, params)
        self.tcex.log.debug(f'status code: {r.status_code}')
        self.tcex.log.trace(f'url: {r.request.url}')
        return r

    @staticmethod
    def is_false(value):
        """checks to see if a string is False"""
        if not value:
            return False
        value = str(value)
        return value.lower() in ['false', '0', 'f', 'n', 'no']

    @staticmethod
    def is_true(value):
        """checks to see if a string is True"""
        if not value:
            return False
        value = str(value)
        return value.lower() in ['true', '1', 't', 'y', 'yes']

    def deleted(
        self, main_type, sub_type, deleted_since=None, owner=None, filters=None, params=None
    ):
        """

        Args:
            owner:
            filters:
            main_type:
            sub_type:
            deleted_since:
            params:

        Return:

        """
        params = params or {}

        if filters and filters.filters:
            params['filters'] = filters.filters_string
        if owner:
            params['owner'] = owner
        if deleted_since:
            params['deletedSince'] = deleted_since

        if not sub_type:
            url = f'/v2/{main_type}/deleted'
        else:
            url = f'/v2/{main_type}/{sub_type}/deleted'

        r = self._get(url, params)

        if not self.success(r):
            err = r.text or r.reason
            self.tcex.handle_error(950, [r.status_code, err, r.url])

        data = r.json().get('data', {}).get('indicator', [])

        yield from data

    def pivot_from_tag(self, target, tag_name, filters=None, owner=None, params=None):
        """

        Args:
            owner:
            filters:
            target:
            tag_name:
            params:

        Return:

        """
        sub_type = target.api_branch
        api_type = target.api_type
        api_entity = target.api_entity
        params = params or {}

        if owner:
            params['owner'] = owner

        if filters and filters.filters:
            params['filters'] = filters.filters_string
        if sub_type:
            url = f'/v2/tags/{tag_name}/{api_type}/{sub_type}'
        else:
            url = f'/v2/tags/{tag_name}/{api_type}/'
        return self._iterate(url, params, api_entity)

    def groups_from_tag(self, group, tag_name, filters=None, owner=None, params=None):
        """

        Args:
            owner:
            group:
            tag_name:
            filters:
            params:

        Return:

        """
        return self.pivot_from_tag(group, tag_name, filters=filters, owner=owner, params=params)

    def indicators_from_tag(self, indicator, tag_name, filters=None, owner=None, params=None):
        """
                Args:
                    owner:
                    indicator:
                    tag_name:
                    filters:
                    params:

                Return:

        """
        params = params or {}

        yield from self.pivot_from_tag(
            indicator, tag_name, filters=filters, owner=owner, params=params
        )

    def victims_from_tag(self, victim, tag_name, filters=None, owner=None, params=None):
        """

        Args:
            owner:
            victim:
            tag_name:
            filters:
            params:

        Return:

        """
        yield from self.pivot_from_tag(
            victim, tag_name, filters=filters, owner=owner, params=params
        )

    def indicator_associations(self, main_type, sub_type, unique_id, owner=None, params=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            params:

        Return:

        """
        params = params or {}
        if owner:
            params['owner'] = owner

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/indicators'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/indicators'

        yield from self._iterate(url, params, 'indicator')

    def group_associations(self, main_type, sub_type, unique_id, owner=None, params=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            params:

        Return:

        """
        params = params or {}
        if owner:
            params['owner'] = owner

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/groups'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/groups'

        yield from self._iterate(url, params, 'group')

    def victim_asset_associations(
        self, main_type, sub_type, unique_id, asset_type=None, owner=None, params=None
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            asset_type:
            params:

        Return:

        """
        params = params or {}

        if owner:
            params['owner'] = owner

        url = f'/v2/{main_type}'
        if sub_type:
            url = f'{url}/{sub_type}'
        url = f'{url}/{unique_id}/victimAssets'
        if asset_type:
            url = f'{url}/{asset_type}'

        yield from self._iterate(url, params, 'victimAsset')

    def indicator_associations_types(
        self,
        main_type,
        sub_type,
        unique_id,
        association_type,
        api_branch=None,
        api_entity=None,
        owner=None,
        params=None,
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            association_type:
            api_branch:
            api_entity:
            params:

        Return:

        """
        params = params or {}
        if owner:
            params['owner'] = owner

        api_branch = api_branch or association_type.api_branch
        api_entity = api_entity or association_type.api_entity
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/indicators/{api_branch}'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/indicators/{api_branch}'

        yield from self._iterate(url, params, api_entity)

    def group_associations_types(
        self,
        main_type,
        sub_type,
        unique_id,
        target,
        api_branch=None,
        api_entity=None,
        owner=None,
        params=None,
    ):
        """
                Args:
                    owner:
                    main_type:
                    sub_type:
                    unique_id:
                    target:
                    api_branch:
                    api_entity:
                    params:

                Return:

        """
        params = params or {}
        if owner:
            params['owner'] = owner

        api_branch = api_branch or target.api_branch
        api_entity = api_entity or target.api_entity

        if target and target.is_task():
            if not sub_type:
                url = f'/v2/{main_type}/{unique_id}/tasks/{api_branch}'
            else:
                url = f'/v2/{main_type}/{sub_type}/{unique_id}/tasks'
        elif not sub_type:
            url = f'/v2/{main_type}/{unique_id}/groups/{api_branch}'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/groups/{api_branch}'

        yield from self._iterate(url, params, api_entity)

    def add_association(
        self,
        main_type,
        sub_type,
        unique_id,
        target_type,
        target_sub_type,
        target_unique_id,
        owner=None,
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            target_type:
            target_sub_type:
            target_unique_id:

        Return:

        """
        return self._association(
            main_type,
            sub_type,
            unique_id,
            target_type,
            target_sub_type,
            target_unique_id,
            owner=owner,
        )

    def delete_association(
        self,
        main_type,
        sub_type,
        unique_id,
        target_type,
        target_sub_type,
        target_unique_id,
        owner=None,
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            target_type:
            target_sub_type:
            target_unique_id:

        Return:

        """
        return self._association(
            main_type,
            sub_type,
            unique_id,
            target_type,
            target_sub_type,
            target_unique_id,
            action='DELETE',
            owner=owner,
        )

    def _association(
        self,
        main_type,
        sub_type,
        unique_id,
        target_type,
        target_sub_type,
        target_unique_id,
        action='ADD',
        owner=None,
    ):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            target_type:
            target_sub_type:
            target_unique_id:
            action:

        Return:

        """
        action = action.upper()
        params = {'owner': owner} if owner else {}
        params['createActivityLog'] = False
        # Typically if victim to victim but other endpoints that are not
        # groups/indicators can exist in the future
        if not sub_type and not target_sub_type:
            url = f'/v2/{main_type}/{unique_id}/{target_type}/{target_unique_id}'
        # Typically if victim to anything else
        elif not sub_type:
            url = f'/v2/{main_type}/{unique_id}/{target_type}/{target_sub_type}/{target_unique_id}'
        # If anything else to victim
        elif not target_sub_type:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/{target_type}/{target_unique_id}'
        else:
            # Everything else
            url = (
                f'/v2/{main_type}/{sub_type}/{unique_id}/{target_type}/{target_sub_type}'
                f'/{target_unique_id}'
            )

        r = None
        params['createActivityLog'] = params.get('createActivityLog') or False
        if action == 'ADD':
            r = self._post(url, {}, params)
        elif action == 'DELETE':
            r = self._delete(url, params)
        else:
            self.tcex.log.error('associations error')

        if r is not None:
            self.tcex.log.debug(f'status code: {r.status_code}')
            self.tcex.log.trace(f'url: {r.request.url}')
        return r

    def victim(self, main_type, sub_type, unique_id, victim_id, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            victim_id:
            params:

        Return:

        """
        params = params or {}
        params['createActivityLog'] = False

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/victims/{victim_id}'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/victims/{victim_id}'

        return self._get(url, params)

    def victims(self, main_type, sub_type, unique_id, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            params:

        Return:

        """
        params = params or {}

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/victims'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/victims'

        yield from self._iterate(url, params, 'victim')

    def file_occurrences(self, main_type, sub_type, unique_id, owner=None):
        """
        Yields all occurrences a File has

        Args:
            main_type:
            sub_type:
            unique_id:
            owner:
        """
        url = f'/v2/{main_type}/{sub_type}/{unique_id}/fileOccurrences'
        params = {}

        if owner:
            params['owner'] = owner

        yield from self._iterate(url, params, 'fileOccurrence')

    def get_file_hash(self, main_type, sub_type, unique_id, hash_type='sha256'):
        """ Gets the hash of a file. """
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/download'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/download'

        if hash_type == 'sha256':
            hashed_file = hashlib.sha256()  # nosec
        elif hash_type == 'sha1':
            hashed_file = hashlib.sha1()  # nosec
        else:
            hashed_file = hashlib.md5()  # nosec

        with self.tcex.session.get(url, stream=True) as r:
            for chunk in r.iter_content(chunk_size=4096):
                if chunk:  # filter out keep-alive new chunks
                    hashed_file.update(chunk)
        return hashed_file

    def get_file_occurrence(self, main_type, sub_type, unique_id, occurrence_id, owner=None):
        """
        Gets a file occurrence given a occurrence_id
        Args:
            main_type:
            sub_type:
            unique_id:
            occurrence_id:
            owner:

        Returns:

        """
        return self.file_occurrence(main_type, sub_type, unique_id, occurrence_id, owner)

    def delete_file_action(self, main_type, sub_type, unique_id, action, target, owner=None):
        """Deletes a file action given a target"""
        url = (
            f'/v2/{main_type}/{sub_type}/{unique_id}/actions/{action}'
            f'/{target.api_type}/{target.api_branch}/{target.unique_id}'
        )
        params = {}

        if owner:
            params['owner'] = owner

        return self._delete(url, params)

    def add_file_action(self, main_type, sub_type, unique_id, action, target, owner=None):
        """Creates a file action given a target"""
        url = (
            f'/v2/{main_type}/{sub_type}/{unique_id}/actions/{action}'
            f'/{target.api_type}/{target.api_branch}/{target.unique_id}'
        )
        params = {}

        if owner:
            params['owner'] = owner

        return self._post(url, {}, params=params)

    def get_file_actions(self, main_type, sub_type, unique_id, action, target, owner=None):
        """Yields a file action given a file action and a target"""
        url = f'/v2/{main_type}/{sub_type}/{unique_id}/actions/{action}/{target.api_type}'

        if target.api_sub_type:
            url = f'{url}/{target.api_branch}'

        params = {}

        if owner:
            params['owner'] = owner

        yield from self._iterate(url, params, target.api_entity)

    def delete_file_occurrence(self, main_type, sub_type, unique_id, occurrence_id, owner=None):
        """
        Deletes a file occurrence given a occurrence_id
        Args:
            main_type:
            sub_type:
            unique_id:
            occurrence_id:
            owner:

        Returns:

        """
        return self.file_occurrence(main_type, sub_type, unique_id, occurrence_id, owner, 'DELETE')

    def file_occurrence(
        self, main_type, sub_type, unique_id, occurrence_id, owner=None, action='GET'
    ):
        """
        Gets a file occurrence given a occurrence_id
        Args:
            main_type:
            sub_type:
            unique_id:
            occurrence_id:
            owner:
            action:

        Returns:

        """
        url = f'/v2/{main_type}/{sub_type}/{unique_id}/fileOccurrences/{occurrence_id}'
        params = {}

        if owner:
            params['owner'] = owner

        if action == 'GET':
            return self._get(url, params)

        if action == 'DELETE':
            return self._delete(url, params)

        return None

    def add_file_occurrence(self, main_type, sub_type, unique_id, name, date, path, owner=None):
        """
        Adds a file occurrence to a File
        Args:
            main_type:
            sub_type:
            unique_id:
            name:
            date:
            path:
            owner:

        Returns:

        """
        url = f'/v2/{main_type}/{sub_type}/{unique_id}/fileOccurrences'
        params = {}
        if owner:
            params['owner'] = owner
        return self._post_json(url, {'fileName': name, 'path': path, 'date': date})

    def tag(self, main_type, sub_type, unique_id, tag, action='GET', owner=None, params=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            tag:
            action:
            params:

        Return:

        """
        params = params or {}

        if owner:
            params['owner'] = owner

        action = action.upper()
        if sub_type:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/tags/{quote(tag)}'
        else:
            url = f'/v2/{main_type}/{unique_id}/tags/{quote(tag)}'
        response = None
        if action == 'ADD':
            response = self._post(url, {}, params)
        elif action == 'DELETE':
            response = self._delete(url, params)
        elif action == 'GET':
            response = self._get(url, params)
        else:
            self.tcex.log.error('_tags error')
        return response

    def add_tag(self, main_type, sub_type, unique_id, tag, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            tag:

        Return:

        """
        return self.tag(main_type, sub_type, unique_id, tag, action='ADD', owner=owner)

    def delete_tag(self, main_type, sub_type, unique_id, tag, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            tag:

        Return:

        """
        return self.tag(main_type, sub_type, unique_id, tag, action='DELETE', owner=owner)

    def get_tag(self, main_type, sub_type, unique_id, tag, owner=None, params=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            tag:
            params:

        Return:

        """
        params = params or {}

        return self.tag(main_type, sub_type, unique_id, tag, owner=owner, params=params)

    def tags(self, main_type, sub_type, unique_id, owner=None, filters=None, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            owner:
            filters:
            params:

        Return:

        """
        params = params or {}

        if owner:
            params['owner'] = owner
        if filters and filters.filters:
            params['filters'] = filters.filters_string
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/tags'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/tags'

        yield from self._iterate(url, params, 'tag')

    def labels(self, main_type, sub_type, unique_id, owner=None, filters=None, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            owner:
            filters:
            params:

        Return:

        """
        params = params or {}

        if owner:
            params['owner'] = owner
        if filters and filters.filters:
            params['filters'] = filters.filters_string
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/securityLabels'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/securityLabels'

        yield from self._iterate(url, params, 'securityLabel')

    def add_label(self, main_type, sub_type, unique_id, label, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            label:

        Return:

        """
        return self.label(main_type, sub_type, unique_id, label, action='ADD', owner=owner)

    def get_label(self, main_type, sub_type, unique_id, label, owner=None, params=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            label:
            params:

        Return:

        """
        params = params or {}

        return self.label(
            main_type, sub_type, unique_id, label, action='GET', owner=owner, params=params
        )

    def delete_label(self, main_type, sub_type, unique_id, label, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            label:

        Return:

        """
        return self.label(main_type, sub_type, unique_id, label, action='DELETE', owner=owner)

    def label(self, main_type, sub_type, unique_id, label, action='ADD', owner=None, params=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            label:
            action:
            params:

        Return:

        """
        params = params or {}

        if owner:
            params['owner'] = owner

        action = action.upper()

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/securityLabels/{quote(label)}'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/securityLabels/{quote(label)}'

        if action == 'ADD':
            return self._post(url, {}, params)

        if action == 'DELETE':
            return self._delete(url, params)

        if action == 'GET':
            return self._get(url, params)

        return None

    def attributes(self, main_type, sub_type, unique_id, owner=None, params=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            params:

        Return:

        """

        params = params or {}

        if owner:
            params['owner'] = owner

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/attributes'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/attributes'

        yield from self._iterate(url, params, 'attribute')

    def attribute(
        self, main_type, sub_type, unique_id, attribute_id, action='GET', owner=None, params=None
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_id:
            action:
            params:

        Return:

        """
        params = params or {}
        if owner:
            params['owner'] = owner
        action = action.upper()
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/attributes/{attribute_id}'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/attributes/{attribute_id}'

        if action == 'GET':
            return self._get(url, params)

        if action == 'DELETE':
            return self._delete(url, params)

        return None

    def get_attribute(self, main_type, sub_type, unique_id, attribute_id, owner=None, params=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_id:
            params:

        Return:

        """
        return self.attribute(
            main_type, sub_type, unique_id, attribute_id, action='GET', owner=owner, params=params
        )

    def delete_attribute(self, main_type, sub_type, unique_id, attribute_id, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_id:

        Return:

        """
        return self.attribute(
            main_type, sub_type, unique_id, attribute_id, action='DELETE', owner=owner
        )

    def add_attribute(
        self,
        main_type,
        sub_type,
        unique_id,
        attribute_type,
        attribute_value,
        source=None,
        displayed=None,
        owner=None,
        params=None,
    ):
        """

        Args:
            displayed:
            source:
            params:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_type:
            attribute_value:

        Return:

        """
        if params is None:
            params = {}
        if owner:
            params['owner'] = owner
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/attributes'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/attributes'

        json = {'type': attribute_type, 'value': attribute_value}

        if source:
            json['source'] = source

        if displayed:
            json['displayed'] = displayed

        return self._post_json(url, json, params)

    def update_attribute(
        self,
        main_type,
        sub_type,
        unique_id,
        attribute_value,
        attribute_id,
        source=None,
        displayed=None,
        owner=None,
        params=None,
    ):
        """

        Args:
            displayed:
            source:
            params:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_type:
            attribute_value:

        Return:

        """
        if params is None:
            params = {}
        if owner:
            params['owner'] = owner
        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/attributes/{attribute_id}'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/attributes/{attribute_id}'

        json = {'value': attribute_value}

        if source:
            json['source'] = source

        if displayed:
            json['displayed'] = displayed

        return self._put_json(url, json, params)

    def attribute_labels(
        self, main_type, sub_type, unique_id, attribute_id, owner=None, params=None
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_id:
            params:

        Return:

        """
        params = params or {}
        if owner:
            params['owner'] = owner

        if not sub_type:
            url = f'/v2/{main_type}/{unique_id}/attributes/{attribute_id}/securityLabels'
        else:
            url = f'/v2/{main_type}/{sub_type}/{unique_id}/attributes/{attribute_id}/securityLabels'

        yield from self._iterate(url, params, 'securityLabel')

    def attribute_label(
        self,
        main_type,
        sub_type,
        unique_id,
        attribute_id,
        label,
        action='GET',
        owner=None,
        params=None,
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_id:
            label:
            action:
            params:

        Return:

        """
        params = params or {}
        if owner:
            params['owner'] = owner
        action = action.upper()

        if not sub_type:
            url = (
                f'/v2/{main_type}/{unique_id}/attributes/{attribute_id}/securityLabels/'
                f'{quote(label)}'
            )
        else:
            url = (
                f'/v2/{main_type}/{sub_type}/{unique_id}/attributes/{attribute_id}/securityLabels/'
                f'{quote(label)}'
            )
        if action == 'ADD':
            return self._post(url, {}, params)

        if action == 'DELETE':
            return self._delete(url, params)

        if action == 'GET':
            return self._get(url, params)

        return None

    def get_attribute_label(
        self, main_type, sub_type, unique_id, attribute_id, label, owner=None, params=None
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_id:
            label:
            params:

        Return:

        """
        return self.attribute_label(
            main_type, sub_type, unique_id, attribute_id, label, owner=owner, params=params
        )

    def delete_attribute_label(
        self, main_type, sub_type, unique_id, attribute_id, label, owner=None
    ):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_id:
            label:

        Return:

        """
        return self.attribute_label(
            main_type, sub_type, unique_id, attribute_id, label, action='DELETE', owner=owner
        )

    def add_attribute_label(self, main_type, sub_type, unique_id, attribute_id, label, owner=None):
        """

        Args:
            owner:
            main_type:
            sub_type:
            unique_id:
            attribute_id:
            label:

        Return:

        """
        return self.attribute_label(
            main_type, sub_type, unique_id, attribute_id, label, action='ADD', owner=owner
        )

    def assignees(self, main_type, unique_id, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            params:

        Return:

        """
        params = params or {}

        url = f'/v2/{main_type}/{unique_id}/assignees'
        yield from self._iterate(url, params, 'user')

    def assignee(self, main_type, unique_id, assignee_id, action='ADD', params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            assignee_id:
            action:
            params:

        Return:

        """
        params = params or {}

        url = f'/v2/{main_type}/{unique_id}/assignees/{assignee_id}'
        if action == 'ADD':
            return self._post(url, {}, params)
        if action == 'GET':
            return self._get(url, params=params)
        if action == 'DELETE':
            return self._delete(url, params)
        return None

    def get_assignee(self, main_type, sub_type, unique_id, assignee_id, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            assignee_id:
            params:

        Return:

        """
        params = params or {}

        return self.assignee(main_type, sub_type, unique_id, assignee_id, params=params)

    def delete_assignee(self, main_type, sub_type, unique_id, assignee_id):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            assignee_id:

        Return:

        """
        return self.assignee(main_type, sub_type, unique_id, assignee_id, 'DELETE')

    def add_assignee(self, main_type, sub_type, unique_id, assignee_id):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            assignee_id:

        Return:

        """
        return self.assignee(main_type, sub_type, unique_id, assignee_id, 'ADD')

    def escalatees(self, main_type, unique_id, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            params:

        Return:

        """
        params = params or {}

        url = f'/v2/{main_type}/{unique_id}/escalatees'
        yield from self._iterate(url, params, 'user')

    def escalatee(self, main_type, unique_id, escalatee_id, action='GET', params=None):
        """

        Args:
            main_type:
            unique_id:
            escalatee_id:
            action:
            params:

        Return:

        """
        params = params or {}

        url = f'/v2/{main_type}/{unique_id}/escalatees/{escalatee_id}'
        if action == 'ADD':
            return self._post(url, {}, params)
        if action == 'GET':
            return self._get(url=url, params=params)
        if action == 'DELETE':
            return self._delete(url=url)
        return None

    def get_escalatee(self, main_type, sub_type, unique_id, escalatee_id, params=None):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            escalatee_id:
            params:

        Return:

        """
        params = params or {}

        return self.escalatee(main_type, sub_type, unique_id, escalatee_id, params=params)

    def delete_escalatee(self, main_type, sub_type, unique_id, escalatee_id):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            escalatee_id:

        Return:

        """
        return self.escalatee(main_type, sub_type, unique_id, escalatee_id, 'DELETE')

    def add_escalatee(self, main_type, sub_type, unique_id, escalatee_id):
        """

        Args:
            main_type:
            sub_type:
            unique_id:
            escalatee_id:

        Return:

        """
        return self.escalatee(main_type, sub_type, unique_id, escalatee_id, 'ADD')

    @staticmethod
    def success(r):
        """

        Args:
            r:

        Return:

        """
        status = True
        if r.ok:
            try:
                if r.json().get('status') != 'Success':
                    status = False
            except Exception:
                status = False
        else:
            status = False
        return status



"""ThreatConnect Case"""
from .api_endpoints import ApiEndpoints
from .common_case_management import CommonCaseManagement
from .common_case_management_collection import CommonCaseManagementCollection
from .filter import Filter
from .tql import TQL


class Cases(CommonCaseManagementCollection):
    """Case Class for Case Management Collection

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        initial_response (dict, optional): Initial data in
            Case Object for Artifact. Defaults to None.
        tql_filters (list, optional): List of TQL filters. Defaults to None.
        params(dict, optional): Dict of the params to be sent while
            retrieving the Case objects.
    """

    def __init__(self, tcex, initial_response=None, tql_filters=None, params=None):
        """Initialize Class properties."""
        super().__init__(
            tcex,
            ApiEndpoints.CASES,
            initial_response=initial_response,
            tql_filters=tql_filters,
            params=params,
        )
        if initial_response:
            for item in initial_response.get('data', []):
                self.added_items.append(Case(tcex, **item))

    def __iter__(self):
        """Object iterator"""
        return self.iterate(initial_response=self.initial_response)

    def entity_map(self, entity):
        """Map a dict to a Artifact.

        Args:
            entity (dict): The Artifact data.

        Returns:
            CaseManagement.Case: A Case Object.
        """
        return Case(self.tcex, **entity)

    @property
    def filter(self):
        """Return instance of FilterCases Object."""
        return FilterCases(ApiEndpoints.CASES, self.tcex, self.tql)


class Case(CommonCaseManagement):
    """Case object for Case Management.

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        artifacts (Artifact, kwargs): a list of Artifacts corresponding to the Case
        assignee (Assignee, kwargs): the user or group Assignee object for the Case
        created_by (User, kwargs): [Read-Only] The **Created By** for the Case.
        date_added (str, kwargs): [Read-Only] The **Date Added** for the Case.
        description (str, kwargs): The **Description** for the Case.
        name (str, kwargs): [Required] The **Name** for the Case.
        notes (Note, kwargs): a list of Notes corresponding to the Case
        owner (str, kwargs): The Org/Owner name for the Case.
        related (Case, kwargs): The **Related** for the Case.
        resolution (str, kwargs): The **Resolution** for the Case.
        severity (str, kwargs): [Required] The **Severity** for the Case.
        status (str, kwargs): [Required] The **Status** for the Case.
        tags (case_management.tag.Tag, kwargs): a list of Tags corresponding to the Case
            (NOTE: Setting this parameter will replace any existing tag(s) with
            the one(s) specified)
        tasks (case_management.task.Task, kwargs): a list of Tasks corresponding to the Case
        user_access (User, kwargs): a list of Users that, when defined, are the only
            ones allowed to view or edit the Case
        workflow_events (WorkflowEvent, kwargs): The **Events** for the Case.
        workflow_template (WorkflowTemplate, kwargs): the Template that the Case is
            populated by.
        xid (str, kwargs): The **Xid** for the Case.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties"""
        super().__init__(tcex, ApiEndpoints.CASES, kwargs)
        self.case_filter = [
            {'keyword': 'caseid', 'operator': TQL.Operator.EQ, 'value': self.id, 'type': 'integer'}
        ]

        self._artifacts = kwargs.get('artifacts', None)
        self._assignee = kwargs.get('assignee', None)
        self._created_by = kwargs.get('created_by', None)
        self._date_added = kwargs.get('date_added', None)
        self._description = kwargs.get('description', None)
        self._name = kwargs.get('name', None)
        self._notes = kwargs.get('notes', None)
        self._owner = kwargs.get('owner', None)
        self._related = kwargs.get('related', None)
        self._resolution = kwargs.get('resolution', None)
        self._severity = kwargs.get('severity', None)
        self._status = kwargs.get('status', None)
        self._tags = kwargs.get('tags', None)
        self._tasks = kwargs.get('tasks', None)
        self._user_access = kwargs.get('user_access', None)
        self._workflow_events = kwargs.get('workflow_events', None)
        self._workflow_template = kwargs.get('workflow_template', None)
        self._xid = kwargs.get('xid', None)

    def add_artifact(self, **kwargs):
        """Add a Artifact to a Case."""
        self.artifacts.add_artifact(self.tcex.cm.artifact(**kwargs))

    def add_note(self, **kwargs):
        """Add a Note to a Case."""
        self.notes.add_note(self.tcex.cm.note(**kwargs))

    def add_tag(self, **kwargs):
        """Add a Tag to a Case."""
        self.tags.add_tag(self.tcex.cm.tag(**kwargs))

    def add_task(self, **kwargs):
        """Add a Task to a Case."""
        self.tasks.add_task(self.tcex.cm.task(**kwargs))

    @property
    def artifacts(self):
        """Return the **Artifacts** for the Case."""
        if self._artifacts is None or isinstance(self._artifacts, dict):
            artifacts = self._artifacts or {}
            self._artifacts = self.tcex.cm.artifacts(
                initial_response=artifacts, tql_filters=self.case_filter
            )
        return self._artifacts

    @artifacts.setter
    def artifacts(self, artifacts):
        """Set the **Artifacts** for the Case."""
        self._artifacts = artifacts

    @property
    def assignee(self):
        """Return the **Assignee** for the Case."""
        if isinstance(self._assignee, dict):
            return self.tcex.cm.assignee(
                type=self._assignee.get('type'), **self._assignee.get('data')
            )
        return self._assignee

    @assignee.setter
    def assignee(self, assignee):
        """Set the **Assignee** for the Case."""
        if isinstance(assignee, dict):
            self._assignee = self.tcex.cm.assignee(
                type=assignee.get('type'), **assignee.get('data')
            )
        self._assignee = assignee

    @property
    def as_entity(self):
        """Return the entity representation of the Case."""
        return {'type': 'Case', 'id': self.id, 'value': self.name}

    @property
    def created_by(self):
        """Return the **Created By** for the Case."""
        if self._created_by:
            return self.tcex.cm.user(**self._created_by)
        return self._created_by

    @property
    def description(self):
        """Return the **Description** for the Case."""
        return self._description

    @description.setter
    def description(self, description):
        """Set the **Description** for the Case."""
        self._description = description

    @property
    def date_added(self):
        """Return the **Date Added** for the Case."""
        return self._date_added

    def entity_mapper(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): An entity dict used to update the Object.
        """
        new_case = Case(self.tcex, **entity)
        self.__dict__.update(new_case.__dict__)

    @property
    def name(self):
        """Return the **Name** for the Case."""
        return self._name

    @name.setter
    def name(self, name):
        """Set the **Name** for the Case."""
        self._name = name

    @property
    def notes(self):
        """Return the **Notes** for the Case."""
        if self._notes is None or isinstance(self._notes, dict):
            notes = self._notes or {}
            self._notes = self.tcex.cm.notes(initial_response=notes, tql_filters=self.case_filter)
        return self._notes

    @notes.setter
    def notes(self, notes):
        """Set the **Notes** for the Case."""
        self._notes = notes

    @property
    def owner(self):
        """Return the **Owner** for the Case."""
        return self._owner

    @property
    def related(self):
        """Return the **Related Cases** for the Case."""
        if self._related:
            return self.tcex.cm.cases(initial_response=self._related, tql_filters=self.case_filter)
        return self._related

    @property
    def resolution(self):
        """Return the **Resolution** for the Case."""
        return self._resolution

    @resolution.setter
    def resolution(self, resolution):
        """Set the **Resolution** for the Case."""
        self._resolution = resolution

    @property
    def severity(self):
        """Return the **Severity** for the Case."""
        return self._severity

    @severity.setter
    def severity(self, severity):
        """Set the **Severity** for the Case."""
        self._severity = severity

    @property
    def status(self):
        """Return the **Status** for the Case."""
        return self._status

    @status.setter
    def status(self, status):
        """Set the **Status** for the Case."""
        self._status = status

    @property
    def tags(self):
        """Return the **Tags** for the Case."""
        if self._tags is None or isinstance(self._tags, dict):
            tags = self._tags or {}
            self._tags = self.tcex.cm.tags(initial_response=tags, tql_filters=self.case_filter)
        return self._tags

    @tags.setter
    def tags(self, tags):
        """Set the **Tags** for the Case."""
        self._tags = tags

    @property
    def tasks(self):
        """Return the **Tasks** for the Case."""
        if self._tasks is None or isinstance(self._tasks, dict):
            tasks = self._tasks or {}
            self._tasks = self.tcex.cm.tasks(initial_response=tasks, tql_filters=self.case_filter)
        return self._tasks

    @tasks.setter
    def tasks(self, tasks):
        """Set the **Tasks** for the Case."""
        self._tasks = tasks

    @property
    def user_access(self):
        """Return the **User Access** for the Case."""
        if self._user_access is None:
            self._user_access = self.tcex.cm.users(users=None)
        elif isinstance(self._user_access, dict):
            self._user_access = self.tcex.cm.users(users=self._user_access.get('data', []))
        return self._user_access

    @user_access.setter
    def user_access(self, user_access):
        """Set the **User Access** for the Case."""
        self._user_access = user_access

    @property
    def workflow_events(self):
        """Return the **Workflow Events** for the Case."""
        if self._workflow_events:
            return self.tcex.cm.workflow_events(
                initial_response=self._workflow_events, tql_filters=self.case_filter
            )
        return self._workflow_events

    @property
    def workflow_template(self):
        """Return the **Workflow Template** for the Case."""
        if self._workflow_template:
            return self.tcex.cm.workflow_template(**self._workflow_template)
        return self._workflow_template

    @property
    def xid(self):
        """Return the **XID** for the Case."""
        return self._xid

    @xid.setter
    def xid(self, xid):
        """Set the **XID** for the Case."""
        self._xid = xid


class FilterCases(Filter):
    """Filter Object for Cases"""

    def created_by(self, operator, created_by):
        """Filter Cases based on **createdBy** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            created_by (str): The account login of the user who created the case.
        """
        self._tql.add_filter('createdBy', operator, created_by, TQL.Type.STRING)

    def created_by_id(self, operator, created_by_id):
        """Filter Cases based on **createdById** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            created_by_id (int): The user ID for the creator of the case.
        """
        self._tql.add_filter('createdById', operator, created_by_id, TQL.Type.INTEGER)

    def date_added(self, operator, date_added):
        """Filter Cases based on **dateAdded** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            date_added (str): The date the case was added to the system.
        """
        self._tql.add_filter('dateAdded', operator, date_added, TQL.Type.STRING)

    def description(self, operator, description):
        """Filter Cases based on **description** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            description (str): The description of the case.
        """
        self._tql.add_filter('description', operator, description, TQL.Type.STRING)

    @property
    def has_artifact(self):
        """Return **FilterArtifacts** for further filtering."""
        from .artifact import FilterArtifacts  # pylint: disable=cyclic-import

        artifacts = FilterArtifacts(ApiEndpoints.ARTIFACTS, self._tcex, TQL())
        self._tql.add_filter('hasArtifact', TQL.Operator.EQ, artifacts, TQL.Type.SUB_QUERY)
        return artifacts

    @property
    def has_note(self):
        """Return **FilterNotes** for further filtering."""
        from .note import FilterNotes  # pylint: disable=cyclic-import

        notes = FilterNotes(ApiEndpoints.NOTES, self._tcex, TQL())
        self._tql.add_filter('hasNote', TQL.Operator.EQ, notes, TQL.Type.SUB_QUERY)
        return notes

    @property
    def has_tag(self):
        """Return **FilterTags** for further filtering."""
        from .tag import FilterTags  # pylint: disable=cyclic-import

        tags = FilterTags(ApiEndpoints.TAGS, self._tcex, TQL())
        self._tql.add_filter('hasTag', TQL.Operator.EQ, tags, TQL.Type.SUB_QUERY)
        return tags

    @property
    def has_task(self):
        """Return **FilterTask** for further filtering."""
        from .task import FilterTasks  # pylint: disable=cyclic-import

        tasks = FilterTasks(ApiEndpoints.TASKS, self._tcex, TQL())
        self._tql.add_filter('hasTask', TQL.Operator.EQ, tasks, TQL.Type.SUB_QUERY)
        return tasks

    def id(self, operator, id):  # pylint: disable=redefined-builtin
        """Filter Cases based on **id** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            id (int): The ID of the case.
        """
        self._tql.add_filter('id', operator, id, TQL.Type.INTEGER)

    def name(self, operator, name):
        """Filter Cases based on **name** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            name (str): The name of the case.
        """
        self._tql.add_filter('name', operator, name, TQL.Type.STRING)

    def owner(self, operator, owner):
        """Filter Cases based on **owner** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            owner (int): The Owner ID for the case.
        """
        self._tql.add_filter('owner', operator, owner, TQL.Type.INTEGER)

    def owner_name(self, operator, owner_name):
        """Filter Cases based on **ownerName** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            owner_name (str): The owner name for the case.
        """
        self._tql.add_filter('ownerName', operator, owner_name, TQL.Type.STRING)

    def resolution(self, operator, resolution):
        """Filter Cases based on **resolution** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            resolution (str): The resolution of the case.
        """
        self._tql.add_filter('resolution', operator, resolution, TQL.Type.STRING)

    def severity(self, operator, severity):
        """Filter Cases based on **severity** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            severity (str): The severity of the case.
        """
        self._tql.add_filter('severity', operator, severity, TQL.Type.STRING)

    def status(self, operator, status):
        """Filter Cases based on **status** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            status (str): The status of the case.
        """
        self._tql.add_filter('status', operator, status, TQL.Type.STRING)

    def tag(self, operator, tag):
        """Filter Cases based on **tag** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            tag (str): The name of a tag applied to a case.
        """
        self._tql.add_filter('tag', operator, tag, TQL.Type.STRING)

    def target_id(self, operator, target_id):
        """Filter Cases based on **targetId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            target_id (int): The assigned user or group ID for the case.
        """
        self._tql.add_filter('targetId', operator, target_id, TQL.Type.INTEGER)

    def target_type(self, operator, target_type):
        """Filter Cases based on **targetType** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            target_type (str): The target type for this case (either User or Group).
        """
        self._tql.add_filter('targetType', operator, target_type, TQL.Type.STRING)

    def xid(self, operator, xid):
        """Filter Cases based on **xid** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            xid (str): The XID of the case.
        """
        self._tql.add_filter('xid', operator, xid, TQL.Type.STRING)



"""TcEx Service Common Module"""
# standard library
import base64
import json
import os
import subprocess
import sys
import time
import uuid
from multiprocessing import Process
from random import randint
from threading import Event, Lock, Thread
from typing import Any, Optional

from ..services import MqttMessageBroker
from .test_case_playbook_common import TestCasePlaybookCommon


class TestCaseServiceCommon(TestCasePlaybookCommon):
    """Service App TestCase Class"""

    _mqtt_client = None
    _message_broker = None
    app_process = None
    client_topic = f'client-topic-{randint(100, 999)}'
    lock = Lock()  # lock for subscription management
    server_topic = f'server-topic-{randint(100, 999)}'
    service_ready = Event()
    service_run_method = 'subprocess'  # run service as subprocess, multiprocess, or thread
    shutdown = False
    shutdown_complete = False
    sleep_after_publish_config = 0.5
    sleep_after_publish_webhook_event = 0.5
    sleep_after_service_start = 5
    sleep_before_delete_config = 2
    sleep_before_shutdown = 0.5
    subscriptions = {}  # dictionary of events, posted to by MID by on_subscribe
    trigger_requests = {}
    trigger_responses = {}

    def _app_callback(self, app):
        """Set app object from run.py callback"""
        self.app = app

    @property
    def default_args(self):
        """Return App default args."""
        args = super().default_args.copy()
        args.update(
            {
                'tc_svc_broker_host': os.getenv('TC_SVC_BROKER_HOST', 'localhost'),
                'tc_svc_broker_port': int(os.getenv('TC_SVC_BROKER_PORT', '1883')),
                'tc_svc_broker_service': os.getenv('TC_SVC_BROKER_SERVICE', 'mqtt'),
                'tc_svc_broker_conn_timeout': int(os.getenv('TC_SVC_BROKER_CONN_TIMEOUT', '15')),
                'tc_svc_broker_token': os.getenv('TC_SVC_BROKER_TOKEN'),
                'tc_svc_client_topic': self.client_topic,
                'tc_svc_server_topic': self.server_topic,
                'tc_svc_hb_timeout_seconds': int(os.getenv('TC_SVC_HB_TIMEOUT_SECONDS', '300')),
            }
        )
        return args

    @property
    def message_broker(self):
        """Return an instance of MqttMessageBroker."""
        if self._message_broker is None:
            self._message_broker = MqttMessageBroker(
                broker_host=self.default_args.get('tc_svc_broker_host'),
                broker_port=self.default_args.get('tc_svc_broker_port'),
                broker_timeout=self.default_args.get('tc_svc_broker_conn_timeout'),
                logger=self.log,
            )
            self._message_broker.register_callbacks()

        return self._message_broker

    @property
    def mqtt_client(self):
        """Return a mqtt client instance."""
        return self.message_broker.client

    def on_message(self, client, userdata, message):  # pylint: disable=unused-argument
        """Handle message broker on_message shutdown command events."""

        try:
            m = json.loads(message.payload)
        except ValueError:
            raise RuntimeError(f'Could not parse API service response JSON. ({message})')

        command = m.get('command').lower()
        if command == 'ready':
            self.service_ready.set()

        trigger_id = str(m.get('triggerId'))
        cmd_type = m.get('type', command).lower()
        key = f'{trigger_id}-{cmd_type}'
        if key in self.trigger_responses:
            self.log.warning(f'Additional response arrived for trigger {key} -- discarded')
        else:
            self.trigger_responses[key] = m
        if self.trigger_requests.get(key) is not None:
            self.trigger_requests.pop(key).set()

    # pylint: disable=unused-argument
    def on_subscribe(self, client, userdata, mid, granted_qos):
        """Handle on subscribe callback after subscription completes"""

        with self.lock:
            event = self.subscriptions.pop(mid)
            if event:
                event.set()

    def publish(self, message: str, topic: Optional[str] = None):
        """Publish message on server channel.

        Args:
            message: The message to send.
            topic: The message broker topic. Defaults to None.
        """
        self.message_broker.publish(message, topic)

    def publish_create_config(self, message: dict) -> dict:
        """Send create config message.

        Args:
            trigger_id (str): The trigger id for the config message.
            message (dict): The entire message with trigger_id and config.

        Returns:
            dict: CreateConfig Acknowledge response data.
        """
        # merge the message config (e.g., optional, required)
        message_config = message.pop('config')
        config = message_config.get('optional', {})
        config.update(message_config.get('required', {}))
        message['config'] = config

        # build config message
        message['apiToken'] = self.tc_token
        message['expireSeconds'] = int(time.time() + 86400)
        message['command'] = 'CreateConfig'
        message['config']['tc_playbook_out_variables'] = self.profile.tc_playbook_out_variables
        message['triggerId'] = message.pop('trigger_id')
        self.log.data('run', 'create config', f'{message["triggerId"]}')
        self.message_broker.publish(json.dumps(message), self.server_topic)

        # create thread wait event
        event = Event()
        key = f'{message["triggerId"]}-createconfig'
        self.trigger_requests[key] = event
        event.wait(10)  # wait for 10 seconds
        # time.sleep(self.sleep_after_publish_config)
        result = self.trigger_responses.pop(key, {})
        self.log.data(
            'run',
            'create config',
            f'{message["triggerId"]} {result.get("status", "No Status")} '
            f'{result.get("message", "No Message")}',
        )
        return dict(status=result.get('status', False), message=result.get('message'))

    def publish_delete_config(self, message):
        """Send delete config message.

        Args:
            message (str): The message coming in on Broker channel
        """
        time.sleep(self.sleep_before_delete_config)
        # using triggerId here instead of trigger_id do to pop in publish_create_config
        config_msg = {'command': 'DeleteConfig', 'triggerId': message.get('triggerId')}
        self.log.data('run', 'delete config', f'{message["triggerId"]}')
        self.message_broker.publish(json.dumps(config_msg), self.server_topic)

    def publish_heartbeat(self):
        """Send heartbeat message to service."""
        shutdown_msg = {'command': 'Heartbeat', 'metric': {}, 'memoryPercent': 0, 'cpuPercent': 0}
        self.message_broker.publish(json.dumps(shutdown_msg), self.server_topic)

    def publish_shutdown(self):
        """Publish shutdown message."""
        config_msg = {'command': 'Shutdown'}
        self.log.data('run', 'service', 'shutdown requested')
        self.message_broker.publish(json.dumps(config_msg), self.server_topic)

    def publish_webhook_event(
        self,
        trigger_id: int,
        body: Optional[Any] = None,
        headers: Optional[list] = None,
        method: Optional[str] = 'GET',
        query_params: Optional[list] = None,
        request_key: Optional[str] = None,
    ):
        """Send create config message.

        Args:
            trigger_id: The trigger ID.
            body: The HTTP request body.
            headers: The HTTP request headers name/value pairs.
            method: The HTTP request method.
            query_params: The HTTP request query param name/value pairs.
            request_key: The current request key.
        """
        body = body or ''
        request_key = request_key or str(uuid.uuid4())
        if isinstance(body, dict):
            body = json.dumps(body)

        body = self.redis_client.hset(
            request_key, 'request.body', base64.b64encode(body.encode('utf-8'))
        )
        event = {
            'command': 'WebhookEvent',
            'method': method,
            'queryParams': query_params or [],
            'headers': headers or [],
            'body': 'request.body',
            'requestKey': request_key,
            'triggerId': trigger_id,
        }

        trigger_id = str(trigger_id)
        self.log.data('run', 'webhook event', trigger_id)
        ready = Event()
        key = f'{trigger_id}-webhookevent'
        self.trigger_requests[key] = ready
        self.message_broker.publish(json.dumps(event), self.server_topic)
        ready.wait(10)  # wait for 10 seconds
        status = 'Complete' if ready.isSet() else 'Timed Out'
        self.trigger_responses.pop(key, {})
        self.log.data('run', 'webhook event', f'{trigger_id} {status}')

    def publish_marshall_webhook_event(
        self,
        trigger_id: int,
        body: Optional[Any] = None,
        headers: Optional[list] = None,
        request_key: Optional[str] = None,
        status_code: Optional[int] = 200,
    ):
        """Send create config message.

        Args:
            trigger_id: The trigger ID.
            body: The HTTP response Body.
            headers: The HTTP response headers name/value pairs.
            request_key: The request key to reply with.
            status_code: The HTTP response status code.
            request_key: The current request key.
        """
        body = body or ''
        request_key = request_key or str(uuid.uuid4())
        if isinstance(body, dict):
            body = json.dumps(body)

        body = self.redis_client.hset(
            request_key, 'request.body', base64.b64encode(body.encode('utf-8'))
        )
        event = {
            'command': 'WebhookMarshallEvent',
            'headers': headers or [],
            'body': 'request.body',
            'requestKey': request_key,
            'triggerId': trigger_id,
            'statusCode': status_code,
        }
        trigger_id = str(trigger_id)
        self.log.data('run', 'webhook marshal', trigger_id)
        ready = Event()
        key = f'{trigger_id}-webhookmarshallevent'
        self.trigger_requests[key] = ready
        self.message_broker.publish(json.dumps(event), self.server_topic)
        ready.wait(10)  # wait for 10 seconds
        status = 'Complete' if ready.isSet() else 'Timed Out'
        self.trigger_responses.pop(key, {})
        self.log.data('run', 'webhook marshal', f'{trigger_id} {status}')

    def run(self):
        """Implement in Child Class"""
        raise NotImplementedError('Child class must implement this method.')

    def run_service(self):
        """Run the micro-service."""
        self.log.data('run', 'service method', self.service_run_method)
        # backup sys.argv
        sys_argv_orig = sys.argv

        # clear sys.argv
        sys.argv = sys.argv[:1]

        # create required .app_params encrypted file. args are set in custom.py
        self.args['tcex_testing_context'] = self.tcex_testing_context
        self.create_config(self.args)

        # run the service App in 1 of 3 ways
        if self.service_run_method == 'subprocess':
            # run the Service App as a subprocess
            self.app_process = subprocess.Popen(['python', 'run.py'])
        elif self.service_run_method == 'thread':
            # run App in a thread
            t = Thread(target=self.run, args=(), daemon=True)
            t.start()
        elif self.service_run_method == 'multiprocess':
            p = Process(target=self.run, args=(), daemon=True)
            p.start()

        # restore sys.argv
        sys.argv = sys_argv_orig

    @classmethod
    def setup_class(cls):
        """Run once before all test cases."""
        super().setup_class()
        cls.args = {}
        cls.service_file = 'SERVICE_STARTED'  # started file flag

        # generate a context used in service.py to write context during fire event
        cls.tcex_testing_context = str(uuid.uuid4())

    def setup_method(self):
        """Run before each test method runs."""
        super().setup_method()
        self.stager.redis.from_dict(self.redis_staging_data)
        self.redis_client = self.tcex.redis_client

        for key in list(self.trigger_requests.keys()):
            del self.trigger_requests[key]

        for key in list(self.trigger_responses.keys()):
            del self.trigger_responses[key]

        self.service_ready.clear()

        # register on_message shutdown monitor
        self.message_broker.add_on_message_callback(
            callback=self.on_message, topics=[self.client_topic]
        )

        # register on_subscribe notification
        self.message_broker.add_on_subscribe_callback(callback=self.on_subscribe)

        self.message_broker.client.loop_start()

        self.subscribe(self.client_topic)

        # only start service if it hasn't been started already base on file flag.
        if not os.path.isfile(self.service_file):
            open(self.service_file, 'w+').close()  # create service started file flag
            self.run_service()

            # start shutdown monitor thread
            t = Thread(target=self.shutdown_monitor, daemon=True)
            t.start()

        self.service_ready.wait(30)
        if not self.service_ready.isSet():
            self.log.data('run', 'service', 'failed to start')
        else:
            self.log.data('run', 'service', 'ready')

    def shutdown_monitor(self):
        """Monitor for shutdown flag."""
        while not self.shutdown:
            time.sleep(0.5)

        # shutdown the App
        self.publish_shutdown()

        # give Service App x seconds to shutdown before terminating
        if self.service_run_method == 'subprocess':
            for _ in range(1, 10):
                time.sleep(0.5)
                if self.app_process.poll() is not None:
                    break
            else:
                self.log.data('run', 'Terminating Process', f'PID: {self.app_process.pid}', 'debug')
                self.app_process.terminate()  # terminate subprocess

        # remove started file flag
        try:
            os.remove(self.service_file)
        except OSError:
            pass

        # set shutdown_complete
        self.shutdown_complete = True

    def stage_data(self, staged_data):
        """Stage the data in the profile."""
        for key, value in list(staged_data.get('redis', {}).items()):
            self.stager.redis.stage(key, value)

    def subscribe(self, topic):
        """Subscribe to a topic and wait for the subscription to complete"""
        event = Event()
        with self.lock:
            # don't try to subscribe outside of lock, since
            # it could come back *before* we wait on it
            _, mid = self.message_broker.client.subscribe(topic)
            self.subscriptions[mid] = event

        event.wait(10)
        if mid in self.subscriptions:  # failed to subscribe
            self.log.error(f'Failed to subscribe to topic {topic}')

    @classmethod
    def teardown_class(cls):
        """Run once before all test cases."""
        super().teardown_class()
        # set shutdown flag for shutdown_monitor and wait until shutdown is done
        cls.shutdown = True
        for _ in range(1, 12):
            if cls.shutdown_complete:
                break
            time.sleep(0.5)

    def teardown_method(self):
        """Run after each test method runs."""
        time.sleep(self.sleep_before_shutdown)
        # run test_case_playbook_common teardown_method
        super().teardown_method()

        self.message_broker.client.loop_stop()

        # clean up tcex testing context after populate_output has run
        self.clear_context(self.tcex_testing_context)



"""Setup for TcEx Module."""
# standard library
import os

# third-party
from setuptools import find_packages, setup

metadata = {}
metadata_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'tcex', '__metadata__.py')
with open(metadata_file, mode='r', encoding='utf-8',) as f:
    exec(f.read(), metadata)  # nosec; pylint: disable=exec-used

if not metadata:
    raise RuntimeError(f'Could not load metadata file ({metadata_file}).')

with open('README.md') as f:
    readme = f.read()

dev_packages = [
    'bandit',
    'black',
    'CommonMark==0.5.5',
    'deepdiff',
    'flake8',
    # isort 5 currently causes issues with pylint
    'isort>=4,<5',
    'mako',
    'pre-commit',
    'pydocstyle',
    'pylint>=2.5.0',
    'pytest',
    'pytest-cov',
    'pytest-html',
    'pytest-xdist>=2.0.0',
    'pyupgrade',
    'recommonmark',
    'reno',
    'sphinx',
    'sphinx-rtd-theme',
]


setup(
    author=metadata['__author__'],
    author_email=metadata['__author_email__'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Security',
    ],
    description=metadata['__description__'],
    download_url=metadata['__download_url__'],
    extras_require={'dev': dev_packages, 'develop': dev_packages, 'development': dev_packages},
    include_package_data=True,
    install_requires=[
        'colorama>=0.3.9',
        'future',
        'hvac>=0.3.0',
        'inflect>=0.2.5',
        'jmespath',
        'jsonschema>=2.6.0',
        'paho-mqtt',
        'parsedatetime',
        'pyaes',
        'python-dateutil>=2.6.1',
        'pytz',
        'redis>=2.10.6',
        'requests>=2.18.4',
        'six>=1.11.0',
        'stdlib-list>=0.6.0',
        'stix2',
        'tzlocal',
        'wrapt',
    ],
    license=metadata['__license__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    name=metadata['__package_name__'],
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_data={'': ['*.json']},
    package_dir={'tcex': 'tcex'},
    project_urls={
        'Documentation': 'https://threatconnect-inc.github.io/tcex/',
        'Source': 'https://github.com/ThreatConnect-Inc/tcex',
    },
    python_requires='>=3.6',
    scripts=[
        'bin/tcinit',
        'bin/tcinit.cmd',
        'bin/tclib',
        'bin/tclib.cmd',
        'bin/tcpackage',
        'bin/tcpackage.cmd',
        'bin/tctest',
        'bin/tctest.cmd',
        'bin/tcvalidate',
        'bin/tcvalidate.cmd',
    ],
    url=metadata['__url__'],
    version=metadata['__version__'],
    zip_safe=True,
)



"""Validators that implement the operators in the operator module."""
# standard library
import operator
from typing import Any, Callable

from .validation_exception import ValidationError


def _operator(
    operator_call: Callable, compare_to: Any, message_posfix: str, allow_none=False,
) -> Callable[..., None]:
    def _validate(value: Any, arg_name: str, label: str):
        """Run validation on input data.

        Args:
            value: The input data to validate.
            arg_name: The name of the input arg.
            label: The label displayed to the user.

        Raises:
            ValidationError: raised on validation failure.
        """
        if not isinstance(value, list):
            value = [value]

        for v in value:
            if allow_none and (v is None or v == ''):
                continue

            if not operator_call(v, compare_to):
                raise ValidationError(f'"{label}" ({arg_name}) {message_posfix}')

    return _validate


def equal_to(compare_to: Any, allow_none=False) -> Callable[..., None]:
    """Validate that an argument is equal to a given value.

    Allowed argument types: String

    Args:
        compare_to (Any): the value to compare the argument value to.
        allow_none (bool): skip any None elements

    Returns:
        A validator function that can be used in the validators argument to @ReadArg.
    """
    formatted_value = f'"{compare_to}"' if isinstance(compare_to, str) else compare_to
    return _operator(operator.eq, compare_to, f'is not equal to {formatted_value}', allow_none)


def less_than(compare_to: Any, allow_none=False) -> Callable[..., None]:
    """Validate that an argument is less than a given value.

    Allowed argument types: String, StringArray

    Args:
        compare_to (Any): the value to compare the argument value to.
        allow_none (bool): skip any None elements

    Returns:
        A validator function that can be used in the validators argument to @ReadArg.
    """
    formatted_value = f'"{compare_to}"' if isinstance(compare_to, str) else compare_to
    return _operator(operator.lt, compare_to, f'is not less than {formatted_value}', allow_none)


def less_than_or_equal(compare_to: Any, allow_none=False) -> Callable[..., None]:
    """Validate that an argument is less than or equal to a given value.

    Allowed argument types: String, StringArray

    Args:
        compare_to (Any): the value to compare the argument value to.
        allow_none (bool): skip any None elements

    Returns:
        A validator function that can be used in the validators argument to @ReadArg.
    """
    formatted_value = f'"{compare_to}"' if isinstance(compare_to, str) else compare_to
    return _operator(
        operator.le, compare_to, f'is not less than or equal to {formatted_value}', allow_none,
    )


def greater_than(compare_to: Any, allow_none=False) -> Callable[..., None]:
    """Validate that an argument is greater than a given value.

    Allowed argument types: String, StringArray

    Args:
        compare_to (Any): the value to compare the argument value to.
        allow_none (bool): skip any None elements

    Returns:
        A validator function that can be used in the validators argument to @ReadArg.
    """
    formatted_value = f'"{compare_to}"' if isinstance(compare_to, str) else compare_to
    return _operator(operator.gt, compare_to, f'is not greater than {formatted_value}', allow_none)


def greater_than_or_equal(compare_to: Any, allow_none=False) -> Callable[..., None]:
    """Validate that an argument is greater than or equal to a given value.

    Allowed argument types: String, StringArray

    Args:
        compare_to (Any): the value to compare the argument value to.
        allow_none (bool): skip any None elements

    Returns:
        A validator function that can be used in the validators argument to @ReadArg.
    """
    formatted_value = f'"{compare_to}"' if isinstance(compare_to, str) else compare_to
    return _operator(
        operator.ge, compare_to, f'is not greater than or equal to {formatted_value}', allow_none,
    )
